# Copyright 2021 Sushil Khanchi <sushilkhanchi97@gmail.com>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
from pathlib import Path
from mercurial import (
    node as node_mod,
    pycompat,
)
import pytest
import grpc
import itertools
# from hgitaly.git import EMPTY_TREE_OID
from hgitaly.license_detector import license_content
from hgitaly.revision import ZERO_SHA
from hgitaly.stub.shared_pb2 import Repository
from hgitaly import stream
from hgitaly.stub.commit_pb2 import (
    FindCommitRequest,
)
from hgitaly.stub.repository_pb2 import (
    CreateRepositoryRequest,
    CreateBundleRequest,
    CreateBundleFromRefListRequest,
    CreateRepositoryFromBundleRequest,
    FindLicenseRequest,
    FindMergeBaseRequest,
    GetArchiveRequest,
    ObjectFormatRequest,
    RemoveRepositoryRequest,
    RepositoryExistsRequest,
    RepositorySizeRequest,
    SearchFilesByContentRequest,
    SearchFilesByNameRequest,
    WriteRefRequest,
)
from hgitaly.stub.shared_pb2 import (
        ObjectFormat,
)
from hgitaly.stub.commit_pb2_grpc import CommitServiceStub
from hgitaly.stub.repository_pb2_grpc import RepositoryServiceStub
from heptapod.testhelpers import (
    LocalRepoWrapper,
    git,
)

from . import skip_comparison_tests
from .comparison import (
    normalize_commit_message,
)
if skip_comparison_tests():  # pragma no cover
    pytestmark = pytest.mark.skip

parametrize = pytest.mark.parametrize

TESTS_DATA_DIR = Path(__file__).parent / 'data'
TIP_TAG_NAME = b'tip'


@parametrize('hg_server', ('hgitaly', 'rhgitaly'))
def test_compare_find_merge_base(gitaly_rhgitaly_comparison, hg_server):
    fixture = gitaly_rhgitaly_comparison
    gitaly_repo = fixture.gitaly_repo
    git_repo = fixture.git_repo
    wrapper = fixture.hg_repo_wrapper

    # repo structure:
    #
    #   o 3 add animal (branch/stable)
    #   |
    #   | 2 add bar
    #   |/
    #   o 1 add zoo
    #   |
    #   o 0 add foo
    #
    gl_branch = b'branch/default'
    sha0 = wrapper.write_commit('foo').hex()
    git_shas = {
        sha0: git_repo.branches()[gl_branch]['sha']
    }
    ctx1 = wrapper.write_commit('zoo')
    sha1 = ctx1.hex()
    git_shas[sha1] = git_repo.branches()[gl_branch]['sha']
    sha2 = wrapper.write_commit('bar').hex()
    git_shas[sha2] = git_repo.branches()[gl_branch]['sha']
    sha3 = wrapper.write_commit('animal', branch='stable', parent=ctx1).hex()
    git_shas[sha3] = git_repo.branches()[b'branch/stable']['sha']
    # commiting a new root, which will test the case when there
    # is no merge_base (gca)
    sha4 = wrapper.commit_file('tut', branch='other',
                               parent=node_mod.nullid).hex()
    git_shas[sha4] = git_repo.branches()[b'branch/other']['sha']

    if hg_server == 'rhgitaly':
        hgitaly_channel = fixture.rhgitaly_channel
    else:
        hgitaly_channel = fixture.hgitaly_channel
    diff_stubs = dict(
        git=RepositoryServiceStub(fixture.gitaly_channel),
        hg=RepositoryServiceStub(hgitaly_channel),
    )

    def do_rpc(vcs, revisions):
        if vcs == 'git':
            revs = [git_shas.get(rev, rev) for rev in revisions]
            revisions = revs

        request = FindMergeBaseRequest(
            repository=gitaly_repo,
            revisions=revisions,
        )

        response = diff_stubs[vcs].FindMergeBase(request)
        base = pycompat.sysbytes(response.base)
        if not base:
            return base
        return base if vcs == 'git' else git_shas[base]

    list_of_interesting_revs = [b'branch/default', b'branch/stable',
                                sha0, sha1, sha4]
    for rev_pair in itertools.product(list_of_interesting_revs, repeat=2):
        assert do_rpc('hg', rev_pair) == do_rpc('git', rev_pair)

    # test with invalid_argument, as it requires minimum 2 revisions
    with pytest.raises(grpc.RpcError) as exc_info_hg:
        do_rpc('hg', [sha0])
    with pytest.raises(grpc.RpcError) as exc_info_git:
        do_rpc('git', [git_shas[sha0]])
    assert exc_info_hg.value.code() == exc_info_git.value.code()
    assert exc_info_hg.value.details() == exc_info_git.value.details()

    sha_not_exists = b'deadnode' * 5
    assert (
        do_rpc('hg', [sha0, sha_not_exists])
        ==
        do_rpc('git', [git_shas[sha0], sha_not_exists])
    )


def test_compare_create_repository(
        gitaly_channel, grpc_channel, server_repos_root):
    rel_path = 'sample_repo'
    default_storage = 'default'
    repo_stubs = dict(
        hg=RepositoryServiceStub(grpc_channel),
        git=RepositoryServiceStub(gitaly_channel)
    )

    def do_rpc(vcs, rel_path, storage=default_storage):
        grpc_repo = Repository(relative_path=rel_path,
                               storage_name=storage)
        request = CreateRepositoryRequest(repository=grpc_repo)
        response = repo_stubs[vcs].CreateRepository(request)
        return response

    hg_rel_path = rel_path + '.hg'
    git_rel_path = rel_path + '.git'
    # actual test
    assert do_rpc('hg', hg_rel_path) == do_rpc('git', git_rel_path)

    # when repo already exists (actually its directory)
    with pytest.raises(grpc.RpcError) as exc_info_git:
        do_rpc('git', git_rel_path)
    with pytest.raises(grpc.RpcError) as exc_info_hg:
        do_rpc('hg', hg_rel_path)
    assert exc_info_hg.value.code() == exc_info_git.value.code()

    # when storage name is invalid
    with pytest.raises(grpc.RpcError) as exc_info_hg:
        do_rpc('hg', rel_path, storage='cargoship')
    with pytest.raises(grpc.RpcError) as exc_info_git:
        do_rpc('git', rel_path, storage='cargoship')
    assert exc_info_hg.value.code() == exc_info_git.value.code()

    # As of 16.8, the broken symlink (pointing to itself) is an INTERNAL
    # error in Gitaly (too many levels of symbolic links),
    # but the more interesting ordinary case
    # of existing directory is still a ALREADY_EXISTS error, so let us
    # test that.
    repo_name = "existing_dir"
    path = (server_repos_root / default_storage / repo_name)
    path.mkdir()
    with pytest.raises(grpc.RpcError) as exc_info_hg:
        do_rpc('hg', repo_name)
    with pytest.raises(grpc.RpcError) as exc_info_git:
        do_rpc('git', repo_name)
    exc_hg, exc_git = exc_info_hg.value, exc_info_git.value
    assert exc_hg.code() == exc_git.code()
    for exc in (exc_hg, exc_git):
        assert 'exists already' in exc.details()


def test_remove_repository(gitaly_comparison, server_repos_root):
    fixture = gitaly_comparison
    grpc_repo = fixture.gitaly_repo
    rpc_helper = fixture.rpc_helper(
        stub_cls=RepositoryServiceStub,
        method_name='RemoveRepository',
        request_cls=RemoveRepositoryRequest,
    )
    assert_compare_errors = rpc_helper.assert_compare_errors

    # unknown storage and missing repo
    assert_compare_errors(same_details=False,
                          repository=Repository(storage_name='unknown',
                                                relative_path='/some/path'))
    assert_compare_errors(
        repository=Repository(storage_name=grpc_repo.storage_name,
                              relative_path='no/such/path'))


@parametrize('hg_server', ('hgitaly', 'rhgitaly'))
def test_repository_exists(gitaly_rhgitaly_comparison,
                           hg_server,
                           server_repos_root):
    fixture = gitaly_rhgitaly_comparison
    grpc_repo = fixture.gitaly_repo
    rpc_helper = fixture.rpc_helper(
        stub_cls=RepositoryServiceStub,
        method_name='RepositoryExists',
        request_cls=RepositoryExistsRequest,
        hg_server=hg_server,
    )
    assert_compare = rpc_helper.assert_compare
    assert_compare_errors = rpc_helper.assert_compare_errors

    assert_compare(repository=grpc_repo)
    assert_compare(
        repository=Repository(storage_name=grpc_repo.storage_name,
                              relative_path='no/such/path'))

    # missing repo *message* (!)
    assert_compare_errors(repository=None, same_details=False)

    # unknown storage
    assert_compare_errors(same_details=False,
                          repository=Repository(storage_name='unknown',
                                                relative_path='/some/path'))


@parametrize('hg_server', ('hgitaly', 'rhgitaly'))
def test_object_format(gitaly_rhgitaly_comparison,
                       hg_server,
                       server_repos_root):
    fixture = gitaly_rhgitaly_comparison
    grpc_repo = fixture.gitaly_repo

    def normalize_unspecified(rpc_helper, resp, vcs='hg', **kw):
        if (
            vcs == 'hg'
            and resp.format == ObjectFormat.OBJECT_FORMAT_UNSPECIFIED
        ):
            # we chose to return UNSPECIFIED, and Gitaly on a repo with default
            # creation options sets and returns SHA1. We might want to revise
            # that later on, but that is not the point of this test.
            resp.format = ObjectFormat.OBJECT_FORMAT_SHA1

    rpc_helper = fixture.rpc_helper(
        stub_cls=RepositoryServiceStub,
        method_name='ObjectFormat',
        request_cls=ObjectFormatRequest,
        hg_server=hg_server,
        normalizer=normalize_unspecified,
    )
    assert_compare = rpc_helper.assert_compare
    assert_compare_errors = rpc_helper.assert_compare_errors

    assert_compare(repository=grpc_repo)
    assert_compare_errors(
        same_details=False,
        repository=Repository(storage_name=grpc_repo.storage_name,
                              relative_path='no/such/path'))

    # missing repo *message* (!)
    assert_compare_errors(repository=None, same_details=False)

    # unknown storage
    assert_compare_errors(same_details=False,
                          repository=Repository(storage_name='unknown',
                                                relative_path='/some/path'))


def test_repository_size(gitaly_comparison,
                         server_repos_root):
    fixture = gitaly_comparison
    grpc_repo = fixture.gitaly_repo
    rpc_helper = fixture.rpc_helper(
        stub_cls=RepositoryServiceStub,
        method_name='RepositorySize',
        request_cls=RepositorySizeRequest,
    )
    # of course Git and Mercurial repository sizes differ, hence
    # we can only compare errors
    assert_compare_errors = rpc_helper.assert_compare_errors

    assert_compare_errors(
        repository=Repository(storage_name=grpc_repo.storage_name,
                              relative_path='no/such/path'))

    # missing repo *message* (!)
    assert_compare_errors(repository=None, same_details=False)

    # unknown storage
    assert_compare_errors(same_details=False,
                          repository=Repository(storage_name='unknown',
                                                relative_path='/some/path'))


@parametrize('hg_server', ('hgitaly', 'rhgitaly'))
def test_has_local_branches(gitaly_rhgitaly_comparison,
                            server_repos_root,
                            hg_server):
    fixture = gitaly_rhgitaly_comparison
    grpc_repo = fixture.gitaly_repo
    rpc_helper = fixture.rpc_helper(
        hg_server=hg_server,
        stub_cls=RepositoryServiceStub,
        method_name='HasLocalBranches',
        request_cls=RemoveRepositoryRequest,
    )
    assert_compare = rpc_helper.assert_compare
    assert_compare_errors = rpc_helper.assert_compare_errors

    assert_compare(repository=grpc_repo)

    # repo does not exist
    assert_compare_errors(
        same_details=False,
        repository=Repository(storage_name=grpc_repo.storage_name,
                              relative_path='no/such/path'))

    # missing repo *message* (!)
    assert_compare_errors(repository=None, same_details=False)

    # unknown storage
    assert_compare_errors(same_details=False,
                          repository=Repository(storage_name='unknown',
                                                relative_path='/some/path'))


def assert_compare_hg_git_created_repos(target_hgrepo, target_gitrepo):
    # assert branches
    br_prefix = b'branch/'
    hgbranches = set(
        [br[0] for br in target_hgrepo.branchmap().iterbranches()])
    gitbranches = set(
        [br[len(br_prefix):] for br in target_gitrepo.branches().keys()])
    assert hgbranches == gitbranches

    # assert tags
    hg_tags = set(target_hgrepo.tags().keys())
    hg_tags.remove(TIP_TAG_NAME)
    git_tags = set(target_gitrepo.tags())
    assert hg_tags == git_tags


def test_compare_create_bundle_and_create_repository_from_bundle(
        gitaly_comparison, tmpdir, server_repos_root):
    default_storage = 'default'
    fixture = gitaly_comparison
    gitaly_repo = fixture.gitaly_repo
    wrapper = fixture.hg_repo_wrapper

    # repo structure:
    #
    #   @ 3 zoo (phase: draft) (amended) (branch:feature)
    #   |
    #   | o 1 bar (phase: public) (tag: v1.2.3)
    #   |/
    #   o 0 foo (phase: public)
    #
    ctx0 = wrapper.commit_file('foo')
    sha0 = ctx0.hex()
    sha1 = wrapper.commit_file('bar').hex()
    wrapper.commit_file('zoo', parent=ctx0, branch='feature')
    wrapper.amend_file('zoo')
    wrapper.set_phase('public', [sha0, sha1])
    wrapper.update(sha1)
    wrapper.command('tag', b'v1.2.3', rev=sha1)
    wrapper.command('gitlab-mirror')

    hg_rel_path = 'target_hg_repo'
    git_rel_path = 'target_git_repo'
    hgrepo_fullpath = server_repos_root / default_storage / hg_rel_path
    gitrepo_fullpath = server_repos_root / default_storage / git_rel_path
    target_repo_msg = dict(
        hg=Repository(relative_path=hg_rel_path,
                      storage_name=default_storage),
        git=Repository(relative_path=git_rel_path,
                       storage_name=default_storage))
    bundle_path = dict(
        hg=tmpdir / 'hg.bundle',
        git=tmpdir / 'git.bundle')
    repo_stub = dict(
        hg=RepositoryServiceStub(fixture.hgitaly_channel),
        git=RepositoryServiceStub(fixture.gitaly_channel))

    def rpc_create_bundle(vcs):
        request = CreateBundleRequest(repository=gitaly_repo)
        response = repo_stub[vcs].CreateBundle(request)
        with open(bundle_path[vcs], 'wb') as fobj:
            for chunk in response:
                fobj.write(chunk.data)

    def rpc_create_repository_from_bundle(vcs):
        with open(bundle_path[vcs], 'rb') as fobj:
            data = fobj.read()
            first_req_data_size = len(data) // 2
            request1 = CreateRepositoryFromBundleRequest(
                repository=target_repo_msg[vcs],
                data=data[:first_req_data_size])
            request2 = CreateRepositoryFromBundleRequest(
                data=data[first_req_data_size:])
            # create an iterator of requests
            request = (req for req in [request1, request2])
            return repo_stub[vcs].CreateRepositoryFromBundle(request)

    # Actual test
    rpc_create_bundle('hg')
    rpc_create_bundle('git')
    rpc_create_repository_from_bundle('hg')
    rpc_create_repository_from_bundle('git')

    target_hgrepo = LocalRepoWrapper.load(hgrepo_fullpath).repo
    target_gitrepo = git.GitRepo(gitrepo_fullpath)

    assert_compare_hg_git_created_repos(target_hgrepo, target_gitrepo)

    # test error case: when repo already exists
    with pytest.raises(grpc.RpcError) as exc_info_git:
        rpc_create_repository_from_bundle('git')
    with pytest.raises(grpc.RpcError) as exc_info_hg:
        rpc_create_repository_from_bundle('hg')
    assert exc_info_hg.value.code() == exc_info_git.value.code()
    assert exc_info_hg.value.details() == exc_info_git.value.details()

    # edge case: empty request
    with pytest.raises(grpc.RpcError) as exc_info_git:
        repo_stub['git'].CreateRepositoryFromBundle(())
    with pytest.raises(grpc.RpcError) as exc_info_hg:
        repo_stub['hg'].CreateRepositoryFromBundle(())
    assert exc_info_hg.value.code() == exc_info_git.value.code()
    assert exc_info_hg.value.details() == exc_info_git.value.details()


def test_create_bundle_from_ref_list(
        gitaly_comparison, tmpdir, server_repos_root):
    default_storage = 'default'
    fixture = gitaly_comparison
    gitaly_repo = fixture.gitaly_repo
    wrapper = fixture.hg_repo_wrapper

    # repo structure:
    #
    #   @ 3 zoo (branch:toy) (phase: draft) (amended)
    #   |
    #   | o 1 bar (branch: animal) (tag: v1.2.3)
    #   |/
    #   o 0 foo (branch: default)
    #
    ctx0 = wrapper.commit_file('foo')
    sha0 = ctx0.hex()
    sha1 = wrapper.commit_file('bar', branch='animal').hex()
    wrapper.commit_file('zoo', parent=ctx0, branch='toys').hex()
    wrapper.amend_file('zoo')
    wrapper.set_phase('public', [sha0, sha1])
    wrapper.update(sha1)
    wrapper.command('tag', b'v1.2.3', rev=sha1)
    wrapper.command('gitlab-mirror')

    def target_repo_path(vcs, bundle_name):
        relative_path = '%s_%s_repo' % (vcs, bundle_name)
        return server_repos_root / default_storage / relative_path

    def target_repo_msg(vcs, bundle_name):
        relative_path = '%s_%s_repo' % (vcs, bundle_name)
        # relative_path = vcs + bundle_name
        return Repository(relative_path=relative_path,
                          storage_name=default_storage)

    def target_repo(vcs, bundle_name):
        path = target_repo_path(vcs, bundle_name)
        if vcs == 'git':
            return git.GitRepo(path)
        return LocalRepoWrapper.load(path).repo.unfiltered()

    def vcs_qualified_bundle_path(vcs, bundle_name):
        return tmpdir / (vcs + bundle_name)

    repo_stub = dict(
        hg=RepositoryServiceStub(fixture.hgitaly_channel),
        git=RepositoryServiceStub(fixture.gitaly_channel),
    )

    def rpc_create_bundle_from_ref_list(
            vcs, bundle_name, refs, without_repository=False):
        bundle_path = vcs_qualified_bundle_path(vcs, bundle_name)

        def get_request_iter(refs):
            first_req = True
            for chunk in stream.split_batches(refs, 2):
                if first_req and not without_repository:
                    first_req = False
                    yield CreateBundleFromRefListRequest(
                        repository=gitaly_repo,
                        patterns=chunk)
                    continue
                yield CreateBundleFromRefListRequest(patterns=chunk)

        request = get_request_iter(refs)
        response = repo_stub[vcs].CreateBundleFromRefList(request)
        with open(bundle_path, 'wb') as fobj:
            for chunk in response:
                fobj.write(chunk.data)

    def rpc_create_repository_from_bundle(vcs, bundle_name):
        bundle_path = vcs_qualified_bundle_path(vcs, bundle_name)

        def get_request_iter(data):
            first_req = True
            for chunk in stream.split_batches(data, 10):
                if first_req:
                    first_req = False
                    yield CreateRepositoryFromBundleRequest(
                        repository=target_repo_msg(vcs, bundle_name),
                        data=chunk)
                    continue
                yield CreateRepositoryFromBundleRequest(data=chunk)

        with open(bundle_path, 'rb') as fobj:
            data = fobj.read()
            request = get_request_iter(data)
            return repo_stub[vcs].CreateRepositoryFromBundle(request)

    def create_bundle(bundle_name, *refs):
        rpc_create_bundle_from_ref_list('hg', bundle_name, refs)
        rpc_create_bundle_from_ref_list('git', bundle_name, refs)

    def create_repository_from_bundle(bundle_name):
        rpc_create_repository_from_bundle('hg', bundle_name)
        rpc_create_repository_from_bundle('git', bundle_name)

    def assert_compare_created_repository_from_bundle(bundle_name):
        target_hgrepo = target_repo('hg', bundle_name)
        target_gitrepo = target_repo('git', bundle_name)
        assert_compare_hg_git_created_repos(target_hgrepo, target_gitrepo)

    # 1) test with all refs
    allrefs_bundle = 'all_refs_bundle'
    create_bundle(
        allrefs_bundle,
        b'refs/heads/branch/animal', b'refs/heads/branch/toys',
        b'refs/heads/branch/default', b'refs/tags/v1.2.3')
    create_repository_from_bundle(allrefs_bundle)
    # test successful repo creation from bundle
    assert_compare_created_repository_from_bundle(allrefs_bundle)

    # 2) test with some refs
    somerefs_bundle = 'some_refs_bundle'
    create_bundle(
        somerefs_bundle,
        b'refs/heads/branch/default', b'refs/heads/branch/toys')
    create_repository_from_bundle(somerefs_bundle)
    # test successful repo creation from bundle
    assert_compare_created_repository_from_bundle(somerefs_bundle)

    # test error case: no repository object in request
    with pytest.raises(grpc.RpcError) as exc_info_git:
        rpc_create_bundle_from_ref_list(
            'git', 'temp_bname', [b'refs/heads/branch/default'],
            without_repository=True)
    with pytest.raises(grpc.RpcError) as exc_info_hg:
        rpc_create_bundle_from_ref_list(
            'hg', 'temp_bname', [b'refs/heads/branch/default'],
            without_repository=True)
    assert exc_info_hg.value.code() == exc_info_git.value.code()
    assert exc_info_hg.value.details() == exc_info_git.value.details()

    # test error case: error in bundle application

    (tmpdir / 'hgbroken-bundle').write("Obviously garbage")
    (tmpdir / 'gitbroken-bundle').write("Garbage, yes, but Git garbage!")
    with pytest.raises(grpc.RpcError) as exc_info_hg:
        rpc_create_repository_from_bundle('hg', 'broken-bundle')
    with pytest.raises(grpc.RpcError) as exc_info_git:
        rpc_create_repository_from_bundle('git', 'broken-bundle')
    assert exc_info_hg.value.code() == exc_info_git.value.code()

    for vcs in ('hg', 'git'):
        assert not target_repo_path(vcs, 'broken-bundle').exists()


def test_search_files_by_name(gitaly_comparison):
    fixture = gitaly_comparison

    wrapper = fixture.hg_repo_wrapper
    ctx0 = wrapper.write_commit('afoo', message="Some foo")
    sub = (wrapper.path / 'sub')
    sub.mkdir()
    (sub / 'bar').write_text('bar content')
    (sub / 'ba2').write_text('ba2 content')
    # TODO OS indep for paths (actually TODO make wrapper.commit easier to
    # use, e.g., check how to make it accept patterns)
    wrapper.commit(rel_paths=['sub/bar', 'sub/ba2'],
                   message="zebar", add_remove=True)

    wrapper.write_commit('animals', message="A branch without subdir",
                         branch='other', parent=ctx0)

    default_rev = b'branch/default'

    rpc_helper = fixture.rpc_helper(stub_cls=RepositoryServiceStub,
                                    method_name='SearchFilesByName',
                                    request_cls=SearchFilesByNameRequest,
                                    request_defaults=dict(
                                        ref=default_rev,
                                        limit=0,
                                        offset=0),
                                    streaming=True,
                                    )
    assert_compare = rpc_helper.assert_compare
    assert_compare_errors = rpc_helper.assert_compare_errors

    # precondition for the test: mirror worked
    assert fixture.git_repo.branch_titles() == {
        default_rev: b"zebar",
        b'branch/other': b"A branch without subdir",
    }

    assert_compare(filter='/bar', query='.')
    assert_compare(filter='^sub/', query='.')
    assert_compare(filter='a', query='sub')
    assert_compare(filter='b?r$', query='sub')

    # without filter
    assert_compare(query='.')
    assert_compare(query='sub')

    # offset and limit (we have 3 matches without filter)
    assert_compare(query='.', limit=2)
    assert_compare(query='.', limit=1, offset=1)  # both sort lexicographically

    # from a different ref
    assert_compare(filter='bar', query='.', ref=b'branch/other')  # no match
    assert_compare(filter='.*', query='.', ref=b'branch/other')

    # query is mandatory
    assert_compare_errors(filter='/ba')

    # problems with regexp
    assert_compare_errors(filter='\\', query='.', same_details=False)
    assert_compare_errors(filter='a' * 1001, query='.')

    # unknown ref gives back an empty list of files, not an error
    assert_compare(ref=b'unknown', query='.')


def test_search_files_by_content(gitaly_comparison):
    fixture = gitaly_comparison

    wrapper = fixture.hg_repo_wrapper
    wrapper.write_commit('afoo', message="Some foo")
    sub = (wrapper.path / 'sub')
    sub.mkdir()
    (sub / 'bar').write_text('line1\nline2\nbar content\nline4\nline5')
    (sub / 'ba2').write_text('line1\nba2 content')
    (sub / 'ba3').write_text('ba3 content\nline2')
    # This one has a Windows endings, and exhibits that `git grep` normalizes
    # to `\n`. Also Git does not interpret the MacOS classic line ending
    # '\r' and we do. In that case, we can claim our response to be more
    # correct and we will not compare it.
    (sub / 'ba4').write_text('l1\r\nl2\nl3\nba4 content\nline6')
    (sub / 'ba5').write_text('m1\nm2\nm3\nm4\nm5\nm6\nba5 content\n')
    (wrapper.path / 'large').write_text('very large content\n' * 3000)
    # TODO OS indep for paths (actually TODO make wrapper.commit easier to
    # use, e.g., check how to make it accept patterns)
    wrapper.commit(rel_paths=['sub/bar', 'sub/ba2', 'sub/ba3',
                              'sub/ba4', 'sub/ba5', 'large'],
                   message="zebar", add_remove=True)

    default_rev = b'branch/default'

    rpc_helper = fixture.rpc_helper(
        stub_cls=RepositoryServiceStub,
        method_name='SearchFilesByContent',
        request_cls=SearchFilesByContentRequest,
        request_defaults=dict(
            ref=default_rev,
        ),
        streaming=True,
        error_details_normalizer=lambda s, vcs: s.lower(),
    )
    assert_compare = rpc_helper.assert_compare
    assert_compare_errors = rpc_helper.assert_compare_errors

    # precondition for the test: mirror worked
    assert fixture.git_repo.branch_titles() == {
        default_rev: b"zebar",
    }
    assert_compare(query='no match for this one')
    assert_compare(query='^bar.c')
    assert_compare(query='^ba2.c')   # only one line before match
    assert_compare(query='^ba3.c')   # only one line after match
    assert_compare(query='^ba4.c')   # more than two lines before match
    assert_compare(query='^very')
    assert_compare(query='^l1|ba4')  # two matches with overlapping context
    assert_compare(query='^m1|ba5')  # two matches with non-overlapping context
    assert_compare(query='ConTent')  # several files and case insensity

    # errors if query is missing
    assert_compare_errors()
    assert_compare_errors(ref=b'topic/default/does-not-exist')
    # unresolvable ref
    assert_compare(ref=b'topic/default/does-not-exist', query='foo')
    # missing repo, ref
    assert_compare_errors(ref=b'')
    assert_compare_errors(repository=None)
    fixture.gitaly_repo.relative_path = 'no/such/repo'
    assert_compare_errors(query='foo', same_details=False)


def test_find_license(gitaly_comparison):
    fixture = gitaly_comparison

    rpc_helper = fixture.rpc_helper(
        stub_cls=RepositoryServiceStub,
        method_name='FindLicense',
        request_cls=FindLicenseRequest,
    )
    rpc_helper.assert_compare()

    wrapper = fixture.hg_repo_wrapper
    wrapper.write_commit('LICENSE', content=license_content('GPL-2'))
    rpc_helper.assert_compare()

    wrapper.write_commit('LICENSE', content="Some garbage")
    rpc_helper.assert_compare()


def test_write_ref(gitaly_comparison, server_repos_root):
    fixture = gitaly_comparison
    rpc_helper = fixture.rpc_helper(
        stub_cls=RepositoryServiceStub,
        method_name='WriteRef',
        request_sha_attrs=['revision', 'old_revision'],
        request_cls=WriteRefRequest,
    )

    def commit_normalizer(rpc_helper, response, **kw):
        if response.HasField('commit'):
            normalize_commit_message(response.commit)

    find_commit_helper = fixture.rpc_helper(
        stub_cls=CommitServiceStub,
        method_name='FindCommit',
        request_cls=FindCommitRequest,
        request_sha_attrs=['revision'],
        response_sha_attrs=['commit.id', 'commit.parent_ids[]'],
        normalizer=commit_normalizer,
    )
    assert_compare_errors = rpc_helper.assert_compare_errors
    assert_compare = rpc_helper.assert_compare

    wrapper = fixture.hg_repo_wrapper
    sha0 = wrapper.write_commit('afoo', message="Some foo").hex()
    sha1 = wrapper.write_commit('afoo', message="Some foo").hex()
    default_rev = b'branch/default'

    # precondition for the test: mirror worked
    assert fixture.git_repo.branch_titles() == {default_rev: b"Some foo"}

    pipeline_1 = b'refs/pipelines/1'
    assert_compare(ref=pipeline_1, revision=default_rev)
    find_commit_helper.assert_compare(revision=pipeline_1)

    # cases with old_revision
    assert_compare(ref=pipeline_1, revision=sha0, old_revision=sha1)
    assert_compare_errors(ref=pipeline_1, revision=default_rev,
                          old_revision=ZERO_SHA)
    assert_compare_errors(ref=pipeline_1, revision=sha0, old_revision=sha1)

    # old_revision is not honoured for symrefs by Gitaly:
    assert_compare(ref=b'HEAD',
                   revision=b"refs/heads/something",
                   old_revision=b"refs/heads/other")

    # HGitaly is stricter than Gitaly with this (notably keep-arounds
    # consistency checks), and we don't want to downgrade that.
    # See also ordinary, non-comparison, tests for more details

    # unknown storage
    fixture.gitaly_repo.storage_name = 'unknown'
    assert_compare_errors(ref=b'refs/pipelines/37', revision=default_rev,
                          same_details=False)


def test_rhgitaly_archive(hgitaly_rhgitaly_comparison):
    fixture = hgitaly_rhgitaly_comparison
    wrapper = fixture.hg_repo_wrapper

    ctx = wrapper.write_commit('foo', content="Foo")
    (wrapper.path / 'sub').mkdir()
    ctx2 = wrapper.write_commit('sub/bar', content="Bar")
    node_str = ctx.hex().decode()
    node2_str = ctx2.hex().decode()

    rpc_helper = fixture.rpc_helper(
        stub_cls=RepositoryServiceStub,
        method_name='GetArchive',
        request_cls=GetArchiveRequest,
        streaming=True,
        request_defaults=dict(
            commit_id=node2_str,
            path=b'',
            format=GetArchiveRequest.Format.Value('TAR'),
            prefix='archive-dir',
        ),
    )

    assert_compare = rpc_helper.assert_compare
    assert_compare_errors = rpc_helper.assert_compare_errors

    assert_compare()
    assert_compare(commit_id=node_str)
    assert_compare(path=b'sub')
    assert_compare_errors(path=b'/etc/passwd')



# Copyright 2020-2023 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later

import grpc
import pytest
import re

from heptapod.gitlab.hooks import (
    GitLabPreReceiveError,
    GitLabPostReceiveError,
)
from hgitaly.stub.repository_pb2 import (
    ApplyGitattributesRequest,
)
from hgitaly.stub.ref_pb2 import FindTagError
from hgitaly.stub.errors_pb2 import ReferenceNotFoundError
from hgitaly.stub.mercurial_operations_pb2 import PublishChangesetError

from hgitaly.stub.repository_pb2_grpc import RepositoryServiceStub

from ..errors import (
    operation_error_treatment,
    structured_abort
)
from ..testing.context import (
    FakeContextAborter as FakeContext,
)

StatusCode = grpc.StatusCode


class FakeLogger:

    def __init__(self):
        self.records = []

    def error(self, fmt, *args):
        self.records.append(('error', fmt % args))


def test_not_implemented(grpc_channel):
    repo_stub = RepositoryServiceStub(grpc_channel)

    with pytest.raises(grpc.RpcError) as exc_info:
        repo_stub.ApplyGitattributes(ApplyGitattributesRequest())
    exc = exc_info.value

    assert exc.code() == grpc.StatusCode.UNIMPLEMENTED
    assert re.search('https://.*/-/issues/1234567', exc.details()) is not None


def test_structured_abort():
    context = FakeContext()

    # example as close as real life as it gets: found with Gitaly Comparison
    # tests in a call for a non-existing tag.
    with pytest.raises(RuntimeError):
        structured_abort(
            context, StatusCode.NOT_FOUND, "tag does not exist",
            FindTagError(
                tag_not_found=ReferenceNotFoundError(
                    reference_name=b"refs/tags/nosuchtag")
            ))
    assert context.code() == StatusCode.NOT_FOUND
    assert context.details() == "tag does not exist"
    trailing = context.trailing_metadata()
    assert len(trailing) == 1
    assert trailing[0] == ('grpc-status-details-bin',
                           b"\x08\x05\x12\x12tag does not exist\x1aB\n'"
                           b"type.googleapis.com/gitaly.FindTagError\x12\x17\n"
                           b"\x15\n\x13refs/tags/nosuchtag")


def test_operation_error_treatment():
    context = FakeContext()
    logger = FakeLogger()
    with pytest.raises(RuntimeError):
        with operation_error_treatment(context, PublishChangesetError, logger):
            raise GitLabPreReceiveError(b"protected")

    assert context.code() == StatusCode.PERMISSION_DENIED
    trailing = context.trailing_metadata()
    assert len(trailing) == 1
    assert trailing[0] == ('grpc-status-details-bin',
                           b'\x08\x07\x12\x0bnot allowed\x1aD\n1type.'
                           b'googleapis.com/hgitaly.PublishChangesetError'
                           b'\x12\x0f\n\r\x12\tprotected\x18\x01')

    # works with `str` messages as well
    context = FakeContext()  # making sure test does not pass by coincidence
    with pytest.raises(RuntimeError):
        with operation_error_treatment(context, PublishChangesetError, logger):
            raise GitLabPreReceiveError("again protected")

    assert context.code() == StatusCode.PERMISSION_DENIED
    trailing = context.trailing_metadata()
    assert len(trailing) == 1
    assert b'again protected' in trailing[0][1]

    with operation_error_treatment(context, PublishChangesetError, logger):
        raise GitLabPostReceiveError(b"pipeline crash")

    # the b'' due to repr() is not pretty, but it is 100% safe
    assert logger.records == [
        ('error',
         "Error in GitLab post-receive hook: b'pipeline crash'")
    ]



# Copyright 2020 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
from contextlib import contextmanager
import pytest
import re
import time
from mercurial.node import nullhex as NULL_HEX
from hgitaly.stub.shared_pb2 import (
    GlobalOptions,
    PaginationParameter,
    Repository,
)
from hgitaly.stub.commit_pb2 import (
    CommitLanguagesRequest,
    CommitIsAncestorRequest,
    FindCommitRequest,
    FindCommitsRequest,
    LastCommitForPathRequest,
    ListCommitsRequest,
    ListCommitsByOidRequest,
    ListCommitsByRefNameRequest,
    ListFilesRequest,
    ListLastCommitsForTreeRequest,
    RawBlameRequest,
)
from hgitaly.stub.commit_pb2_grpc import CommitServiceStub
from google.protobuf.timestamp_pb2 import Timestamp

from . import skip_comparison_tests
from .comparison import (
    normalize_commit_message,
)
if skip_comparison_tests():  # pragma no cover
    pytestmark = pytest.mark.skip

parametrize = pytest.mark.parametrize


@parametrize('hg_server', ('hgitaly', 'rhgitaly'))
def test_compare_last_commit_for(gitaly_rhgitaly_comparison, hg_server):
    fixture = gitaly_rhgitaly_comparison
    git_repo = fixture.git_repo

    wrapper = fixture.hg_repo_wrapper
    ctx0 = wrapper.write_commit('foo', message="Some foo")
    git_shas = {
        ctx0.hex(): git_repo.branches()[b'branch/default']['sha'],
    }

    sub = (wrapper.path / 'sub')
    sub.mkdir()
    subdir = (sub / 'dir')
    subdir.mkdir()
    (sub / 'bar').write_text('bar content')
    (sub / 'ba2').write_text('ba2 content')
    (subdir / 'bar').write_text('bar content')
    (subdir / 'ba2').write_text('ba2 content')
    # TODO OS indep for paths (actually TODO make wrapper.commit easier to
    # use, e.g., check how to make it accept patterns)
    ctx1 = wrapper.commit(rel_paths=['sub/bar', 'sub/ba2',
                                     'sub/dir/bar', 'sub/dir/ba2'],
                          message="ze\nbar", add_remove=True)
    git_shas[ctx1.hex()] = git_repo.branches()[b'branch/default']['sha']
    ctx2 = wrapper.write_commit('sub/bar', message='default head')
    ctx3 = wrapper.write_commit('foo', parent=ctx1, branch='other',
                                message='other head')

    # mirror worked
    git_branches = git_repo.branches()
    assert set(git_branches) == {b'branch/default', b'branch/other'}

    def response_ignores(rpc_helper, responses, **kw):
        for resp in responses:
            for commit_for_tree in resp.commits:
                normalize_commit_message(commit_for_tree.commit)

    rpc_helper = fixture.rpc_helper(stub_cls=CommitServiceStub,
                                    hg_server=hg_server,
                                    method_name='ListLastCommitsForTree',
                                    streaming=True,
                                    request_cls=ListLastCommitsForTreeRequest,
                                    request_defaults=dict(limit=1000),
                                    request_sha_attrs=['revision'],
                                    response_sha_attrs=[
                                        'commits[].commit.id',
                                        'commits[].commit.parent_ids[]',
                                        ],
                                    normalizer=response_ignores,
                                    )
    assert_compare = rpc_helper.assert_compare
    assert_compare_errors = rpc_helper.assert_compare_errors

    for path in (b'sub/dir', b'sub/dir/', b'', b'.', b'/', b'./',
                 b'sub', b'sub/', b'foo'):
        for rev in ('branch/default', 'branch/other', ctx2.hex(), ctx3.hex()):
            assert_compare(revision=rev, path=path)

    for pattern in (b'fo*', b's*b', b'sub/*/bar'):
        assert_compare(revision='branch/default', path=pattern)

    # offset just enough to be reached by subdirs
    assert_compare(revision='branch/default', path=b'sub', offset=1)
    # offset is after subdirs, there are two files, we'll get just one
    assert_compare(revision='branch/default', path=b'sub/', offset=2)

    # limit at subdirs
    assert_compare(revision='branch/default', path=b'sub', limit=1)
    # limit excludes one of the two files
    assert_compare(revision='branch/default', path=b'sub', offset=1, limit=1)

    # for a bunch of assertions that aren't about revision nor path
    common_args = dict(revision=ctx2.hex(), path=b'')
    assert_compare(limit=0, **common_args)
    assert_compare_errors(limit=-1, **common_args)
    assert_compare_errors(limit=10, offset=-1, **common_args)

    # error won't be due to invalidity as a SHA, but because commit doesn't
    # exist (let's not depend on Gitaly accepting symbolic revisions, here)
    assert_compare_errors(revision=b'be0123ef' * 5, path=b'sub')
    for path in (b'foo', b'sub/'):
        assert_compare_errors(revision=NULL_HEX, path=path)

    assert_compare(revision=ctx2.hex(), path=b'no-such-dir/')

    def lcfp_norm(rpc_helper, response, **kw):
        normalize_commit_message(response.commit)

    rpc_helper = fixture.rpc_helper(stub_cls=CommitServiceStub,
                                    hg_server=hg_server,
                                    method_name='LastCommitForPath',
                                    request_cls=LastCommitForPathRequest,
                                    request_sha_attrs=['revision'],
                                    response_sha_attrs=[
                                        'commit.id',
                                        'commit.parent_ids[]',
                                        ],
                                    normalizer=lcfp_norm,
                                    )

    assert_compare = rpc_helper.assert_compare
    assert_compare_errors = rpc_helper.assert_compare_errors

    for path in (b'sub/dir', b'sub/dir/', b'', b'.', b'/', b'./',
                 b'sub', b'sub/', b'foo'):
        for rev in (b'branch/default', b'branch/other',
                    ctx2.hex(), ctx3.hex()):
            assert_compare(revision=rev, path=path)

    # when requesting on null revision, the error message from RHGitaly
    # is about a repository corruption because it uses internally
    # the same `RevlogError` (`InvalidRevision`) as for missed lookups
    # (which are typically an inconsistency between GitLab state and the
    # repository).
    assert_compare_errors(revision=NULL_HEX, path=b'foo', same_details=False)
    # case when revision does not exist
    for path in (b'', b'foo'):
        # as of Gitaly 16.8, the Gitaly details are
        # 'panic: runtime error: invalid memory address or \
        #    nil pointer dereference'
        # (used to be exit code 128, clearly rewritten from subprocess
        # into native Golang. However it would be over-ridiculous to
        # mimic that one.
        assert_compare_errors(revision=b'be0123ef' * 5, path=path,
                              same_details=False)

    # case when file does not exist or does not match
    assert_compare(revision=b'branch/default', path=b'no-such-file')

    # pathspecs with wildcards by default
    for pathspec in (b'f*o', b'sub/*/bar', b'sub/*2'):
        assert_compare(revision=b'branch/default', path=pathspec)

    # literal pathspecs
    literal_opts = GlobalOptions(literal_pathspecs=True)
    for path in (b'f*o', b'sub', b'sub/'):
        assert_compare(revision=b'branch/default',
                       path=path,
                       global_options=literal_opts,
                       )


def test_compare_list_last_commits_for_tree_rhgitaly_chunked(rhgitaly):
    fixture = rhgitaly

    wrapper = fixture.hg_repo_wrapper
    repo_root = wrapper.path
    ctx0 = wrapper.write_commit('foo', message="Some foo")
    filenames = [b'bar%03d' % i for i in range(100)]
    for fname in filenames:
        (repo_root / fname.decode('ascii')).write_text("Some bar")
    wrapper.commit(rel_paths=filenames,
                   message=f"Added {len(filenames)} files",
                   add_remove=True)

    rpc_helper = fixture.rpc_helper(stub_cls=CommitServiceStub,
                                    method_name='ListLastCommitsForTree',
                                    streaming=True,
                                    request_cls=ListLastCommitsForTreeRequest,
                                    # TODO not really applied, fix that
                                    request_defaults=dict(limit=1000),
                                    )
    resp = rpc_helper.rpc('rhgitaly', revision=b'branch/default', limit=1000)
    assert len(resp) == 3
    resp_offset = rpc_helper.rpc('rhgitaly', revision=b'branch/default',
                                 offset=1, limit=1000)
    assert len(resp_offset) == 2
    foo_result = resp[-1].commits[-1]
    assert foo_result.path_bytes == b'foo'
    assert foo_result.commit.id == ctx0.hex().decode('ascii')


def test_compare_raw_blame(gitaly_comparison):
    fixture = gitaly_comparison

    wrapper = fixture.hg_repo_wrapper
    ctx0 = wrapper.commit_file('foo',
                               content='second_line\n'
                                       'third_line\n')
    ctx1 = wrapper.commit_file('foo',
                               content='first_line\n'
                                       'second_line\n'
                                       'third_line\n'
                                       'forth_line\n')

    RAW_BLAME_LINE_REGEXP = re.compile(br'(\w{40}) (\d+) (\d+)')

    def convert_chunk(rpc_helper, chunk, vcs):
        lines = chunk.splitlines(True)
        final = []
        for line in lines:
            hash_line = RAW_BLAME_LINE_REGEXP.match(line)
            if hash_line is not None:
                hash_id = hash_line.group(1)
                if vcs == 'hg':
                    hash_id = rpc_helper.hg2git(hash_id)
                line_no = hash_line.group(2)
                old_line_no = hash_line.group(2)
                final.append((hash_id, line_no, old_line_no))
            elif line.startswith(b'\t'):
                final.append(line)
        return final

    def normalizer(rpc_helper, responses, vcs=None):
        for i, chunk in enumerate(responses):
            responses[i] = convert_chunk(rpc_helper, chunk.data, vcs)

    rpc_helper = fixture.rpc_helper(
        stub_cls=CommitServiceStub,
        method_name='RawBlame',
        request_cls=RawBlameRequest,
        request_sha_attrs=['revision'],
        streaming=True,
        normalizer=normalizer,
    )

    rpc_helper.assert_compare(revision=ctx0.hex(), path=b'foo')
    rpc_helper.assert_compare(revision=ctx1.hex(), path=b'foo')
    rpc_helper.assert_compare(revision=ctx1.hex(), path=b'foo', range=b'2,3')
    rpc_helper.assert_compare(revision=ctx1.hex(), path=b'foo', range=b'2,3')
    rpc_helper.assert_compare(revision=ctx1.hex(), path=b'foo', range=b'2,4')
    # specifiying an end of range greater than the number of lines is not
    # an error.
    rpc_helper.assert_compare(revision=ctx1.hex(), path=b'foo', range=b'2,100')

    # error cases
    rpc_helper.assert_compare_errors(revision=ctx1.hex(), path=b'')
    rpc_helper.assert_compare_errors(revision=ctx1.hex(), path=b'unknown')
    rpc_helper.assert_compare_errors(revision=ctx1.hex(), path=b'foo',
                                     range=b'1000,1001')


@parametrize('hg_server', ('hgitaly', 'rhgitaly'))
def test_compare_commit_is_ancestor(gitaly_rhgitaly_comparison, hg_server):
    fixture = gitaly_rhgitaly_comparison
    wrapper = fixture.hg_repo_wrapper
    # TODO copied from hgitaly.tests.common, refactor so that the latter
    # can accept an existing repo wrapper.
    chgs = {}
    base = chgs['base'] = wrapper.write_commit('foo', message='Base')
    chgs['default'] = wrapper.write_commit('foo', message='Head of default')
    chgs['other_base'] = wrapper.write_commit(
        'foo', message='Start other', branch='other', parent=base)
    chgs['other'] = wrapper.write_commit('foo', message='Other wild',
                                         branch='other')
    chgs['top1'] = wrapper.write_commit('foo', message='Topic first',
                                        branch='default', topic='zzetop',
                                        parent=base)
    chgs['top2'] = wrapper.write_commit('foo', message='Topic head')

    rpc_helper = fixture.rpc_helper(
        hg_server=hg_server,
        stub_cls=CommitServiceStub,
        method_name='CommitIsAncestor',
        request_cls=CommitIsAncestorRequest,
        request_sha_attrs=['ancestor_id', 'child_id'],
        )
    for ancestor, child in (('base', 'other'),
                            ('other', 'top1'),
                            ('top1', 'top2')):
        rpc_helper.assert_compare(ancestor_id=chgs[ancestor].hex().decode(),
                                  child_id=chgs[child].hex().decode())


def test_compare_list_files(gitaly_comparison):
    fixture = gitaly_comparison
    git_repo = fixture.git_repo

    wrapper = fixture.hg_repo_wrapper
    ctx0 = wrapper.write_commit('foo', message="Some foo")

    sub = (wrapper.path / 'sub')
    sub.mkdir()
    subdir = (sub / 'dir')
    subdir.mkdir()
    (sub / 'bar').write_text('bar content')
    (sub / 'ba2').write_text('ba2 content')
    (subdir / 'bar').write_text('bar content')
    (subdir / 'ba2').write_text('ba2 content')
    # TODO OS indep for paths (actually TODO make wrapper.commit easier to
    # use, e.g., check how to make it accept patterns)
    ctx1 = wrapper.commit(rel_paths=['sub/bar', 'sub/ba2',
                                     'sub/dir/bar', 'sub/dir/ba2'],
                          message="zebar", add_remove=True)
    ctx2 = wrapper.write_commit('sub/bar', message='default head')
    ctx3 = wrapper.write_commit('zoo', parent=ctx0, branch='other',
                                message='other head')

    # mirror worked
    git_branches = git_repo.branches()
    assert set(git_branches) == {b'branch/default', b'branch/other'}

    rpc_helper = fixture.rpc_helper(
        stub_cls=CommitServiceStub,
        method_name='ListFiles',
        request_cls=ListFilesRequest,
        streaming=True,
        request_sha_attrs=['revision'],
    )

    not_exists = b'65face65' * 5
    for rev in [ctx0.hex(), ctx1.hex(), ctx2.hex(), ctx3.hex(),
                not_exists, b'branch/default', b'branch/other']:
        rpc_helper.assert_compare(revision=rev)


@parametrize('hg_server', ('hgitaly', 'rhgitaly'))
def test_compare_find_commit(gitaly_rhgitaly_comparison, hg_server):
    fixture = gitaly_rhgitaly_comparison
    wrapper = fixture.hg_repo_wrapper

    def normalizer(rpc_helper, response, **kw):
        if response.HasField('commit'):
            normalize_commit_message(response.commit)

    rpc_helper = fixture.rpc_helper(
        hg_server=hg_server,
        stub_cls=CommitServiceStub,
        method_name='FindCommit',
        request_cls=FindCommitRequest,
        request_sha_attrs=['revision'],
        response_sha_attrs=['commit.id', 'commit.parent_ids[]'],
        normalizer=normalizer,
        )

    assert_compare = rpc_helper.assert_compare

    assert_compare(revision=b'HEAD')
    fixture.invalidate()  # for the hg->git map

    ctx0 = wrapper.commit_file('foo')
    wrapper.command('tag', b'start-tag', rev=ctx0.hex())
    ctx1 = wrapper.commit_file('foo', topic='sampletop')
    mr_ref_path, _ = fixture.write_special_ref(b'merge-requests/2/train',
                                               ctx1.hex())

    assert_compare(revision=NULL_HEX)
    assert_compare(revision=ctx0.hex())
    assert_compare(revision=b'dead' * 10)  # non resolvable full node
    assert_compare(revision=b'branch/default')
    assert_compare(revision=b'refs/heads/branch/default')
    assert_compare(revision=b'topic/default/sampletop')
    assert_compare(revision=b'refs/heads/topic/default/sampletop')
    assert_compare(revision=b'refs/heads')
    assert_compare(revision=b'start-tag')
    assert_compare(revision=b'refs/tags/start-tag')
    assert_compare(revision=b'refs/tags/unknown')
    assert_compare(revision=b'HEAD')
    assert_compare(revision=mr_ref_path)
    hg_ka_ref_path, git_ka_ref_path = fixture.create_keep_around(ctx1.hex())
    # assert_compare() not being able to convert the keep-around ref path from
    # Mercurial to Git on the fly, we need to go lower level
    hg_resp = rpc_helper.rpc('hg', revision=hg_ka_ref_path)
    git_resp = rpc_helper.rpc('git', revision=git_ka_ref_path)
    rpc_helper.normalize_responses(hg_resp, git_resp)
    assert hg_resp == git_resp

    # collision between branch and tag (test validity corroborated by checking
    # agreement also on the tag ref full path)
    wrapper.command('tag', b'tagbook', rev=ctx1.hex())
    wrapper.command('bookmark', b'tagbook', rev=ctx0.hex())
    wrapper.command('tag', b'branch/default', rev=ctx0.hex())
    wrapper.command('gitlab-mirror')
    fixture.invalidate()  # for the hg->git map
    assert_compare(revision=b"refs/tags/tagbook")
    assert_compare(revision=b"tagbook")
    assert_compare(revision=b"refs/tags/branch/default")
    assert_compare(revision=b"branch/default")

    # collision between tag and node id (full form and shortened)
    # Notice how we do *not* run gitlab-mirror, as we need different tags on
    # both sides. Also no sense using assert_compare() in that case either
    # What matters here is that Gitaly and HGitaly behave identically,
    # but for the sake of completeness, as of this writing, in Gitaly,
    # tags have precedence over shortened commit ids, but not on full
    # commit ids (same with a command-line Git, for what it's worth).
    git_repo = fixture.git_repo
    hg2git = rpc_helper.hg2git
    hg_sha0, hg_sha1 = ctx0.hex(), ctx1.hex()
    git_sha0, git_sha1 = hg2git(hg_sha0), hg2git(hg_sha1)

    for hg_sha_tag, git_sha_tag in ((hg_sha0, git_sha0),
                                    (hg_sha0[:10], git_sha0[:10])):
        wrapper.command('tag', hg_sha_tag, rev=hg_sha1)
        wrapper.command('gitlab-mirror')
        git_repo.write_ref(b'refs/tags/' + git_sha_tag, git_sha1)

        assert rpc_helper.hg2git(
            rpc_helper.rpc('hg', revision=hg_sha_tag).commit.id.encode()
        ) == rpc_helper.rpc('git', revision=git_sha_tag).commit.id.encode()

    # error cases
    assert_compare_errors = rpc_helper.assert_compare_errors
    assert_compare_errors()  # no revision
    assert_compare_errors(repository=None,
                          revision=b'HEAD',
                          same_details=False)
    assert_compare_errors(repository=None,  # and no revision
                          same_details=False)
    fixture.gitaly_repo.relative_path = 'unknown/repo'
    assert_compare_errors(revision=b'HEAD', same_details=False)


@parametrize('hg_server', ('hgitaly', 'rhgitaly'))
def test_compare_count_find_commits(gitaly_rhgitaly_comparison, hg_server):
    fixture = gitaly_rhgitaly_comparison
    wrapper = fixture.hg_repo_wrapper

    def normalizer(rpc_helper, responses, **kw):
        # Sorting is for the special cases where, we have two
        # commits diverging and Git order the commits arbitrarily
        # for e.g.
        #
        #  B
        #  |  C          Here, if selecting from bottom to top, order
        #  | /           can be: (A, B, C) or (A, C, B)
        #  A
        #
        # We actually sort inside each chunk instead of the whole, but
        # that should be enough.
        if rpc_helper.sorted:
            for chunk in responses:
                chunk.commits.sort(key=lambda c: c.id)

        for chunk in responses:
            for commit in chunk.commits:
                normalize_commit_message(commit)

    rpc_helper = fixture.rpc_helper(
        hg_server=hg_server,
        stub_cls=CommitServiceStub,
        method_name='FindCommits',
        request_cls=FindCommitsRequest,
        streaming=True,
        request_defaults=dict(limit=10),
        request_sha_attrs=['revision'],
        response_sha_attrs=['commits[].id', 'commits[].parent_ids[]'],
        normalizer=normalizer,
        )
    rpc_helper.sorted = False

    @contextmanager
    def sorted_comparison():
        orig = rpc_helper.sorted
        rpc_helper.sorted = True
        yield
        rpc_helper.sorted = orig

    assert_compare = rpc_helper.assert_compare
    assert_compare_errors = rpc_helper.assert_compare_errors

    assert_compare_errors(revision=b'HEAD')
    fixture.invalidate()  # for the hg->git map

    # set_default_gitlab_branch(wrapper.repo, b'branch/default')
    # prepare repo as:
    #
    #   @    4 (branch/default) merge with stable
    #   |\
    #   | o  3 creates 'animal' (branch/stable)
    #   | |
    #   o |  2 rename 'foo' to 'zoo' (user: testuser)
    #   |/
    #   | 1 changes 'foo' (topic: sampletop)
    #   |/
    #   o  0  creates 'foo'
    #

    ctx0 = wrapper.commit_file('foo')
    ctx1 = wrapper.commit_file('foo', topic='sampletop')
    wrapper.update(ctx0.rev())
    wrapper.command(b'mv', wrapper.repo.root + b'/foo',
                    wrapper.repo.root + b'/zoo')
    ctx2 = wrapper.commit([b'foo', b'zoo'], message=b"rename foo to zoo")
    ts = int(time.time())
    ctx3 = wrapper.write_commit('animals', branch='stable', parent=ctx0,
                                utc_timestamp=ts+10,
                                user='testuser <testuser@heptapod.test>')
    wrapper.update(2)
    ctx4 = wrapper.merge_commit(ctx3, message=b'merge with stable',
                                utc_timestamp=ts+20)

    # when `revision` is provided as <revspec>
    with sorted_comparison():
        all_revs = [ctx0.hex(), ctx1.hex(), ctx2.hex(), ctx3.hex(), ctx4.hex()]
        for range_str in (b'..', b'...'):
            for r1 in all_revs:
                for r2 in all_revs:
                    assert_compare(revision=r1 + range_str + r2)

    # when `revision` is provided as a ref to a single commit
    refs = [b'', ctx0.hex(), b'topic/default/sampletop', ctx2.hex(),
            b'branch/stable', b'branch/default', b'HEAD']
    for ref in refs:
        rpc_helper.assert_compare(revision=ref)

    # with `include_shortstat` option
    for ref in refs:
        rpc_helper.assert_compare(revision=ref,
                                  include_shortstat=True)

    # with `path` and `follow` options
    test_paths = [
        # Note: we are not including [b'foo'] here, because of a special case:
        # in a rename-cset (foo -> zoo), Git consider the cset but Hg doesn't,
        # as 'foo' is not present in rename-cset.
        [b'zoo'],
        [b'foo', b'zoo'],
    ]
    for follow in [True, False]:
        for paths in test_paths:
            if len(paths) > 1:
                # In Git, 'follow' doesn't work with multiple paths
                follow = False
            rpc_helper.assert_compare(paths=paths, follow=follow)

    # with simple options
    with sorted_comparison():
        assert_compare(all=True)
    assert_compare(author=b'testuser')
    assert_compare(skip_merges=True)

    # with pagination options
    for limit in range(0, 5):
        for offset in range(0, 5):
            assert_compare(offset=offset, limit=limit)
    assert_compare(order=FindCommitsRequest.Order.TOPO)

    # with `after` and `before` options for dates
    date1, date2 = Timestamp(), Timestamp()
    date1.FromSeconds(ts+10)
    date2.FromSeconds(ts+20)
    for date in [date1, date2]:
        assert_compare(after=date)
        assert_compare(before=date)
        assert_compare(before=date, after=date)

    # with `include_referenced_by` option
    mr_ref_path, _ = fixture.write_special_ref(b'merge-requests/2/train',
                                               ctx2.hex())
    wrapper.command('tag', b'release', rev=ctx3.hex())
    all_revs.append(wrapper.repo[b'.'].hex())  # will be HEAD
    wrapper.command('gitlab-mirror')
    fixture.invalidate()  # will need to normalize the tagging changeset itself

    for patterns in ([b'refs/heads'],
                     [b'refs/heads/'],
                     [b'refs/heads/branch/'],
                     [b'refs/tags'],
                     [b'refs/tags/'],
                     [b'refs/tags/rel'],
                     [b'refs/merge-requests'],
                     [b'refs/heads', b'refs/tags'],  # matching is logical OR
                     [b'HEAD'],
                     ):
        for rev in all_revs:
            rpc_helper.assert_compare(revision=rev,
                                      include_referenced_by=patterns)

    # This looks to be simple matching by segment prefixes,
    # but git-log(1) isn't very explicit about the type of matching it is
    # performing (Gitaly uses `git log --decorate-ref` for this as of v15.9):
    #        --decorate-refs=<pattern>, --decorate-refs-exclude=<pattern>
    #            For each candidate reference, do not use it for decoration if
    #            it matches any patterns given to --decorate-refs-exclude or if
    #            it doesn’t match any of the patterns given to --decorate-refs.
    for patterns in ([b'refs/h*ds/'],  # glob
                     [b'refs/tags/rel'],  # general prefixing
                     [b'refs/h..ds/'],  # regexp
                     ):
        rpc_helper.assert_compare(revision=b'branch/default',
                                  include_referenced_by=patterns)

    #
    # Reusing fixture for CountCommits
    #

    rpc_helper = fixture.rpc_helper(
        hg_server=hg_server,
        stub_cls=CommitServiceStub,
        method_name='CountCommits',
        request_cls=FindCommitsRequest,
        request_sha_attrs=['revision'],
    )

    assert_compare = rpc_helper.assert_compare
    assert_compare_errors = rpc_helper.assert_compare_errors

    # when `revision` is provided as <revspec>
    all_revs = [ctx0.hex(), ctx1.hex(), ctx2.hex(), ctx3.hex(), ctx4.hex()]
    for range_str in (b'..', ):  # TODO symdiff not implemented
        for r1 in all_revs:
            for r2 in all_revs:
                assert_compare(revision=r1 + range_str + r2)

    # when `revision` is provided as a ref to a single commit
    refs = [ctx0.hex(), b'topic/default/sampletop', ctx2.hex(),
            b'branch/stable', b'branch/default', b'HEAD']
    for ref in refs:
        rpc_helper.assert_compare(revision=ref)

    assert_compare_errors()


def test_compare_list_commits(gitaly_comparison):
    fixture = gitaly_comparison

    wrapper = fixture.hg_repo_wrapper
    # set_default_gitlab_branch(wrapper.repo, b'branch/default')
    # prepare repo as:
    #
    #   @    4 (branch/default) merge with stable
    #   |\
    #   | o  3 creates 'animal' (branch/stable)
    #   | |
    #   o |  2 rename 'foo' to 'zoo' (user: testuser)
    #   |/
    #   | 1 changes 'foo' (topic: sampletop)
    #   |/
    #   o  0  creates 'foo'
    #

    ctx0 = wrapper.commit_file('foo',
                               message="Imagine it to be bar\n\n"
                               "There is a Match, a quote ' and "
                               r"even a \ backslash")
    ctx1 = wrapper.commit_file('foo', topic='sampletop')
    wrapper.update(ctx0.rev())
    wrapper.command(b'mv', wrapper.repo.root + b'/foo',
                    wrapper.repo.root + b'/zoo')
    ts = int(time.time())
    ctx2 = wrapper.commit([b'foo', b'zoo'],
                          message=b"rename foo to zoo",
                          utc_timestamp=ts - 10)
    ctx3 = wrapper.write_commit('animals', branch='stable', parent=ctx0,
                                utc_timestamp=ts + 10,
                                user='testuser <testuser@heptapod.test')
    wrapper.update(2)
    ctx4 = wrapper.merge_commit(ctx3, message=b'merge with stable',
                                utc_timestamp=ts+20)

    def normalizer(rpc_helper, responses, **kw):
        for chunk in responses:
            for commit in chunk.commits:
                normalize_commit_message(commit)

    rpc_helper = fixture.rpc_helper(
        stub_cls=CommitServiceStub,
        method_name='ListCommits',
        request_cls=ListCommitsRequest,
        streaming=True,
        request_defaults=dict(
            pagination_params=PaginationParameter(limit=10)),
        request_sha_attrs=['revision'],
        response_sha_attrs=['commits[].id', 'commits[].parent_ids[]'],
        normalizer=normalizer,
        )
    rpc_helper.sorted = False

    def request_kwargs_to_git(hg_kwargs):
        """Replace Mercurial SHAs by their Git counterparts.

        The format of the ``revisions`` parameter is too specific to
        be provided directly by :class:`RpcHelper`
        """
        git_kwargs = hg_kwargs.copy()
        revisions = hg_kwargs.get('revisions')
        if revisions is None:
            return git_kwargs

        git_kwargs['revisions'] = git_revisions = []
        for rev in revisions:
            if rev.startswith(b'^'):
                git_rev = b'^' + rpc_helper.revspec_to_git(rev[1:])
            else:
                git_rev = rpc_helper.revspec_to_git(rev)
            git_revisions.append(git_rev)

        pagination = hg_kwargs.get('pagination_params')
        if pagination is not None and pagination.page_token:
            git_kwargs['pagination_params'] = PaginationParameter(
                limit=pagination.limit,
                page_token=rpc_helper.revspec_to_git(pagination.page_token)
            )

        return git_kwargs

    rpc_helper.request_kwargs_to_git = request_kwargs_to_git

    assert_compare = rpc_helper.assert_compare
    assert_compare_errors = rpc_helper.assert_compare_errors

    def caret(ctx):
        return b'^' + ctx.hex()

    assert_compare(revisions=[ctx4.hex(), caret(ctx1)])
    assert_compare(revisions=[ctx4.hex(), caret(ctx1)], reverse=True)
    # interpretation of several exclusions
    assert_compare(revisions=[ctx4.hex(), caret(ctx1), caret(ctx2)])

    # no result
    assert_compare(revisions=[ctx4.hex(), caret(ctx4)])

    # orderings
    #
    # Comparison is limited because Mercurial orderings don't exactly
    # match Git's. See docstring of the `ListCommit` method for details.
    # Notably we can't compare the date ordering (or we would cheat by
    # using special cases where they coincide, which is worse than no test)
    Order = ListCommitsRequest.Order
    assert_compare(revisions=[ctx4.hex()], order=Order.TOPO)

    # commit message patterns
    #
    # In requests matching multiple commits, we have to force ordering
    # because the default Git and Mercurial orderings do not give the same
    assert_compare(revisions=[ctx4.hex(), caret(ctx1)])
    assert_compare(revisions=[ctx4.hex(), caret(ctx1)], reverse=True)
    assert_compare(revisions=[b'branch/default'],
                   commit_message_patterns=[b'zoo'])
    assert_compare(revisions=[b'branch/default'],
                   commit_message_patterns=[br"\\"])
    # default ordering is different on this query, hence we force it
    assert_compare(revisions=[b'branch/default'],
                   commit_message_patterns=[b'zoo', b'body'],
                   order=Order.TOPO)
    assert_compare(revisions=[b'branch/default'],
                   commit_message_patterns=[b'match'])
    assert_compare(revisions=[b'branch/default'],
                   commit_message_patterns=[b'match'],
                   ignore_case=True)
    # limit and page token
    assert_compare(revisions=[ctx4.hex()],
                   order=Order.TOPO,
                   pagination_params=PaginationParameter(limit=2))

    # starting over *after* the emission of ctx2 does not mean
    # completing the just-tested limit request, which stopped right before
    # emitting ctx2 (in case someone inspects the inner values and is puzzled
    # not to see all ancestors of ctx4).
    assert_compare(
        revisions=[ctx4.hex()],
        order=Order.TOPO,
        pagination_params=PaginationParameter(limit=10,
                                              page_token=ctx2.hex()))

    # unknown revision
    assert_compare_errors(revisions=[b'1234' * 10], same_details=False)

    # invalid calls
    assert_compare_errors(order=Order.TOPO)
    assert_compare_errors(revisions=[b'branch/default'],
                          commit_message_patterns=[b'[]'],  # invalid regexp
                          same_details=False)

    assert_compare_errors(revisions=[ctx4.hex()],
                          repository=Repository(storage_name='unknown',
                                                relative_path='/no/matter'),
                          same_details=False)
    assert_compare_errors(revisions=[ctx4.hex()],
                          repository=None,
                          same_details=False)


@parametrize('hg_server', ('hgitaly', 'rhgitaly'))
def test_compare_list_commits_by_oid(gitaly_rhgitaly_comparison, hg_server):
    fixture = gitaly_rhgitaly_comparison
    wrapper = fixture.hg_repo_wrapper

    def normalizer(rpc_helper, responses, **kw):
        for chunk in responses:
            for commit in chunk.commits:
                normalize_commit_message(commit)

    def chunk_concatenator(responses):
        return [c for r in responses for c in r.commits]

    def chunked_fields_remover(response):
        """Empties the 'commits' chunked field

        That leaves no 'small' fields to compare, this is an edge case,
        but it's still ok.
        """
        del response.commits[:]

    rpc_helper = fixture.rpc_helper(
        hg_server=hg_server,
        stub_cls=CommitServiceStub,
        method_name='ListCommitsByOid',
        request_cls=ListCommitsByOidRequest,
        streaming=True,
        request_sha_attrs=['oid'],
        response_sha_attrs=['commits[].id', 'commits[].parent_ids[]'],
        normalizer=normalizer,
        chunks_concatenator=chunk_concatenator,
        chunked_fields_remover=chunked_fields_remover,
        )

    hexes = [
        wrapper.commit_file(
            'foo',
            message=f'foo{x}\n\n'
            'Putting some content to make Gitaly cut\n' * 200
        ).hex().decode('ascii')
        for x in range(110)
    ]
    # one chunk
    rpc_helper.assert_compare(oid=hexes[:10])

    # several chunks
    rpc_helper.assert_compare_aggregated(oid=hexes, compare_first_chunks=False)

    # special cases
    rpc_helper.assert_compare(oid=['0123dead4567cafe0000' * 2])  # no match
    rpc_helper.assert_compare(oid=['not-hexadecimal'])
    # a wrong argument does not interfere with correct ones
    rpc_helper.assert_compare(oid=['not-hexadecimal', hexes[0]])
    # NULL_NODE gets ignored
    rpc_helper.assert_compare(oid=[NULL_HEX])
    rpc_helper.assert_compare(oid=[NULL_HEX, hexes[0]])


@parametrize('hg_server', ('hgitaly', 'rhgitaly'))
def test_compare_list_commits_by_ref_name(gitaly_rhgitaly_comparison,
                                          hg_server):
    fixture = gitaly_rhgitaly_comparison
    wrapper = fixture.hg_repo_wrapper

    def normalizer(rpc_helper, responses, vcs=None):
        for chunk in responses:
            for commit_ref in chunk.commit_refs:
                normalize_commit_message(commit_ref.commit)
                commit_ref.ref_name = rpc_helper.hg2git(  # call w/ direct SHA
                    rpc_helper.normalize_keep_around(
                        commit_ref.ref_name, vcs=vcs)
                )

    rpc_helper = fixture.rpc_helper(
        hg_server=hg_server,
        stub_cls=CommitServiceStub,
        method_name='ListCommitsByRefName',
        request_cls=ListCommitsByRefNameRequest,
        streaming=True,
        response_sha_attrs=[
            'commit_refs[].commit.id',
            'commit_refs[].commit.parent_ids[]'
        ],
        normalizer=normalizer,
    )

    def request_kwargs_to_git(hg_kwargs):
        git_kwargs = hg_kwargs.copy()
        ref_names = hg_kwargs.get('ref_names')
        if ref_names is None:
            return git_kwargs
        git_kwargs['ref_names'] = [
            rpc_helper.hg2git(  # for direct SHA
                rpc_helper.normalize_keep_around(name, vcs='hg')
            )
            for name in ref_names
        ]
        return git_kwargs

    rpc_helper.request_kwargs_to_git = request_kwargs_to_git

    assert_compare = rpc_helper.assert_compare

    # branches
    hg_branches = {
        b'br%02d' % i: wrapper.commit_file('foo', branch='br%02d' % i).hex()
        for i in range(3)
    }
    assert_compare()
    assert_compare(ref_names=[b'refs/heads/branch/' + branch
                              for branch in hg_branches.keys()])

    # implicit GitLab branch notation
    assert_compare(ref_names=[b'branch/br01'])

    # unknown refs are ignored
    unknown_ref = b'refs/unknown'
    assert_compare(ref_names=[b'branch/br01', unknown_ref])
    assert_compare(ref_names=[unknown_ref])
    assert_compare(ref_names=[unknown_ref, b'refs/heads/branch/br02'])

    # with tags
    wrapper.command('tag', b'v3.1', rev=b'br01')
    wrapper.command('gitlab-mirror')
    fixture.invalidate()
    assert_compare(ref_names=[b'refs/heads/tags/v3.1'])
    assert_compare(ref_names=[b'v3.1'])

    # with special refs
    fixture.write_special_ref(b'pipeline/13', hg_branches[b'br02'])
    assert_compare(ref_names=[b'refs/heads/pipelines/13'])

    # with a keep-around
    hg_sha = hg_branches[b'br01']
    fixture.create_keep_around(hg_sha)
    rpc_helper.assert_compare(ref_names=[b'refs/keep-around/' + hg_sha])

    # direct SHA
    rpc_helper.assert_compare(ref_names=[hg_sha])


def test_compare_commit_languages(gitaly_comparison):
    fixture = gitaly_comparison
    wrapper = fixture.hg_repo_wrapper

    rpc_helper = fixture.rpc_helper(
        stub_cls=CommitServiceStub,
        method_name='CommitLanguages',
        request_cls=CommitLanguagesRequest,
        request_sha_attrs=['oid'],
    )
    assert_compare = rpc_helper.assert_compare

    # a Python file with valid content, including a blank line
    wrapper.commit_file('foo.py',
                        content="import sys\n\nprint(sys.version)\n",
                        message="Some Python",
                        )
    # a Ruby file with valid content, with a comment
    # despite Ruby being after Python in lexicographical order,
    # it should come first, being the one with the most bytes.
    wrapper.commit_file('bar.rb',
                        content=("# frozen_string_literal: true\n"
                                 "\n"
                                 "module Bar\n"
                                 "def f(x)\n"
                                 "  x+1\n"
                                 ),
                        message="Some Ruby",
                        )
    assert_compare()

    # GitHub Linguist restricts itself to some language types, namely, from
    # blob_helper.rb:
    #
    #     DETECTABLE_TYPES = [:programming, :markup].freeze
    #
    # it turns out that COBOL is a programming language without any associated
    # color in the `languages.json` file.
    wrapper.commit_file('truc.cob')
    assert_compare()

    # error cases are hard to reproduce permanently in Gitaly



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: shared.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from . import lint_pb2 as lint__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0cshared.proto\x12\x06gitaly\x1a\x1fgoogle/protobuf/timestamp.proto\x1a\nlint.proto\"\xc3\x01\n\nRepository\x12\x14\n\x0cstorage_name\x18\x02 \x01(\t\x12\x15\n\rrelative_path\x18\x03 \x01(\t\x12\x1c\n\x14git_object_directory\x18\x04 \x01(\t\x12(\n git_alternate_object_directories\x18\x05 \x03(\t\x12\x15\n\rgl_repository\x18\x06 \x01(\t\x12\x17\n\x0fgl_project_path\x18\x08 \x01(\tJ\x04\x08\x01\x10\x02J\x04\x08\x07\x10\x08R\x04path\"+\n\rCommitTrailer\x12\x0b\n\x03key\x18\x01 \x01(\x0c\x12\r\n\x05value\x18\x02 \x01(\x0c\"M\n\x0e\x43ommitStatInfo\x12\x11\n\tadditions\x18\x01 \x01(\x05\x12\x11\n\tdeletions\x18\x02 \x01(\x05\x12\x15\n\rchanged_files\x18\x03 \x01(\x05\"\xeb\x02\n\tGitCommit\x12\n\n\x02id\x18\x01 \x01(\t\x12\x0f\n\x07subject\x18\x02 \x01(\x0c\x12\x0c\n\x04\x62ody\x18\x03 \x01(\x0c\x12$\n\x06\x61uthor\x18\x04 \x01(\x0b\x32\x14.gitaly.CommitAuthor\x12\'\n\tcommitter\x18\x05 \x01(\x0b\x32\x14.gitaly.CommitAuthor\x12\x12\n\nparent_ids\x18\x06 \x03(\t\x12\x11\n\tbody_size\x18\x07 \x01(\x03\x12-\n\x0esignature_type\x18\x08 \x01(\x0e\x32\x15.gitaly.SignatureType\x12\x0f\n\x07tree_id\x18\t \x01(\t\x12\'\n\x08trailers\x18\n \x03(\x0b\x32\x15.gitaly.CommitTrailer\x12+\n\x0bshort_stats\x18\x0b \x01(\x0b\x32\x16.gitaly.CommitStatInfo\x12\x15\n\rreferenced_by\x18\x0c \x03(\x0c\x12\x10\n\x08\x65ncoding\x18\r \x01(\t\"g\n\x0c\x43ommitAuthor\x12\x0c\n\x04name\x18\x01 \x01(\x0c\x12\r\n\x05\x65mail\x18\x02 \x01(\x0c\x12(\n\x04\x64\x61te\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x10\n\x08timezone\x18\x04 \x01(\x0c\"\x1b\n\nExitStatus\x12\r\n\x05value\x18\x01 \x01(\x05\"@\n\x06\x42ranch\x12\x0c\n\x04name\x18\x01 \x01(\x0c\x12(\n\rtarget_commit\x18\x02 \x01(\x0b\x32\x11.gitaly.GitCommit\"\xc5\x01\n\x03Tag\x12\x0c\n\x04name\x18\x01 \x01(\x0c\x12\n\n\x02id\x18\x02 \x01(\t\x12(\n\rtarget_commit\x18\x03 \x01(\x0b\x32\x11.gitaly.GitCommit\x12\x0f\n\x07message\x18\x04 \x01(\x0c\x12\x14\n\x0cmessage_size\x18\x05 \x01(\x03\x12$\n\x06tagger\x18\x06 \x01(\x0b\x32\x14.gitaly.CommitAuthor\x12-\n\x0esignature_type\x18\x07 \x01(\x0e\x32\x15.gitaly.SignatureType\"Y\n\x04User\x12\r\n\x05gl_id\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\x0c\x12\r\n\x05\x65mail\x18\x03 \x01(\x0c\x12\x13\n\x0bgl_username\x18\x04 \x01(\t\x12\x10\n\x08timezone\x18\x05 \x01(\t\":\n\nObjectPool\x12,\n\nrepository\x18\x01 \x01(\x0b\x32\x12.gitaly.RepositoryB\x04\x90\xc6,\x01\"8\n\x13PaginationParameter\x12\x12\n\npage_token\x18\x01 \x01(\t\x12\r\n\x05limit\x18\x02 \x01(\x05\"\'\n\x10PaginationCursor\x12\x13\n\x0bnext_cursor\x18\x01 \x01(\t\"*\n\rGlobalOptions\x12\x19\n\x11literal_pathspecs\x18\x01 \x01(\x08*B\n\nObjectType\x12\x0b\n\x07UNKNOWN\x10\x00\x12\n\n\x06\x43OMMIT\x10\x01\x12\x08\n\x04\x42LOB\x10\x02\x12\x08\n\x04TREE\x10\x03\x12\x07\n\x03TAG\x10\x04*_\n\x0cObjectFormat\x12\x1d\n\x19OBJECT_FORMAT_UNSPECIFIED\x10\x00\x12\x16\n\x12OBJECT_FORMAT_SHA1\x10\x01\x12\x18\n\x14OBJECT_FORMAT_SHA256\x10\x02*5\n\rSignatureType\x12\x08\n\x04NONE\x10\x00\x12\x07\n\x03PGP\x10\x01\x12\x08\n\x04X509\x10\x02\x12\x07\n\x03SSH\x10\x03*.\n\rSortDirection\x12\r\n\tASCENDING\x10\x00\x12\x0e\n\nDESCENDING\x10\x01\x42\x34Z2gitlab.com/gitlab-org/gitaly/v16/proto/go/gitalypbb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'shared_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'Z2gitlab.com/gitlab-org/gitaly/v16/proto/go/gitalypb'
  _OBJECTPOOL.fields_by_name['repository']._options = None
  _OBJECTPOOL.fields_by_name['repository']._serialized_options = b'\220\306,\001'
  _globals['_OBJECTTYPE']._serialized_start=1451
  _globals['_OBJECTTYPE']._serialized_end=1517
  _globals['_OBJECTFORMAT']._serialized_start=1519
  _globals['_OBJECTFORMAT']._serialized_end=1614
  _globals['_SIGNATURETYPE']._serialized_start=1616
  _globals['_SIGNATURETYPE']._serialized_end=1669
  _globals['_SORTDIRECTION']._serialized_start=1671
  _globals['_SORTDIRECTION']._serialized_end=1717
  _globals['_REPOSITORY']._serialized_start=70
  _globals['_REPOSITORY']._serialized_end=265
  _globals['_COMMITTRAILER']._serialized_start=267
  _globals['_COMMITTRAILER']._serialized_end=310
  _globals['_COMMITSTATINFO']._serialized_start=312
  _globals['_COMMITSTATINFO']._serialized_end=389
  _globals['_GITCOMMIT']._serialized_start=392
  _globals['_GITCOMMIT']._serialized_end=755
  _globals['_COMMITAUTHOR']._serialized_start=757
  _globals['_COMMITAUTHOR']._serialized_end=860
  _globals['_EXITSTATUS']._serialized_start=862
  _globals['_EXITSTATUS']._serialized_end=889
  _globals['_BRANCH']._serialized_start=891
  _globals['_BRANCH']._serialized_end=955
  _globals['_TAG']._serialized_start=958
  _globals['_TAG']._serialized_end=1155
  _globals['_USER']._serialized_start=1157
  _globals['_USER']._serialized_end=1246
  _globals['_OBJECTPOOL']._serialized_start=1248
  _globals['_OBJECTPOOL']._serialized_end=1306
  _globals['_PAGINATIONPARAMETER']._serialized_start=1308
  _globals['_PAGINATIONPARAMETER']._serialized_end=1364
  _globals['_PAGINATIONCURSOR']._serialized_start=1366
  _globals['_PAGINATIONCURSOR']._serialized_end=1405
  _globals['_GLOBALOPTIONS']._serialized_start=1407
  _globals['_GLOBALOPTIONS']._serialized_end=1449
# @@protoc_insertion_point(module_scope)



# Copyright 2021 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later

import io
import logging

from grpc import StatusCode

from mercurial import (
    error,
)

from ..stub.shared_pb2 import (
    ObjectType,
)
from ..stub.blob_pb2 import (
    GetBlobRequest,
    GetBlobResponse,
    GetBlobsRequest,
    GetBlobsResponse,
)
from ..stub.blob_pb2_grpc import BlobServiceServicer

from ..oid import (
    blob_oid,
    extract_blob_oid,
)
from ..file_context import git_perms
from ..revision import gitlab_revision_changeset
from ..servicer import HGitalyServicer
from ..stream import WRITE_BUFFER_SIZE

base_logger = logging.getLogger(__name__)


class BlobServicer(BlobServiceServicer, HGitalyServicer):
    """BlobService implementation.

    The ordering of methods in this source file is the same as in the proto
    file.
    """

    STATUS_CODE_STORAGE_NOT_FOUND = StatusCode.INVALID_ARGUMENT

    def GetBlob(self, request: GetBlobRequest,
                context) -> GetBlobResponse:
        if not request.oid:
            context.abort(StatusCode.INVALID_ARGUMENT, "empty Oid")

        # TODO return Unavailable for readLimit = 0, as Gitaly does
        repo = self.load_repo(request.repository, context).unfiltered()
        chgs_id, path = extract_blob_oid(repo, request.oid)
        try:
            changeset = repo[chgs_id.encode()]
        except error.RepoLookupError:
            yield GetBlobResponse()
            return

        try:
            filectx = changeset.filectx(path)
        except error.ManifestLookupError:
            yield GetBlobResponse()
            return

        size = filectx.size()
        data = filectx.data()
        if request.limit != -1:
            data = data[:request.limit]

        for chunk, is_first in iter_blob_chunks(data):
            resp_dict = dict(data=chunk)
            if is_first:
                resp_dict.update(size=size, oid=request.oid)
            yield GetBlobResponse(**resp_dict)

    def GetBlobs(self, request: GetBlobsRequest, context) -> GetBlobsResponse:
        revision_paths = [(rev_path.revision, rev_path.path)
                          for rev_path in request.revision_paths]

        repo = self.load_repo(request.repository, context).unfiltered()
        limit = request.limit

        for rev, path in revision_paths:
            # TODO it's probably an oversight on upstream side that revision
            # is `str` and not `bytes`, but then, what would be the encoding?
            changeset = gitlab_revision_changeset(repo,
                                                  rev.encode('utf-8', 'ignore')
                                                  )
            if changeset is None:
                filectx = None
            else:
                try:
                    filectx = changeset.filectx(path)
                except error.ManifestLookupError:
                    filectx = None

            if changeset is None or filectx is None:
                # missing file is represented by oid=''
                # (see comment in blob.proto)
                yield GetBlobsResponse(path=path, revision=rev)
                continue

            filectx = changeset.filectx(path)
            oid = blob_oid(repo, changeset.hex().decode(), path)

            size = filectx.size()
            data = filectx.data()
            if limit != -1:
                data = data[:limit]

            for chunk, first in iter_blob_chunks(data):
                # is_submodule will be False, because that's the default
                # should at least be documented somewhere though
                dict_resp = dict(data=chunk)

                if first:
                    dict_resp.update(size=size,
                                     oid=oid,
                                     path=path,
                                     revision=rev,
                                     type=ObjectType.BLOB,
                                     mode=git_perms(filectx),
                                     )

                yield GetBlobsResponse(**dict_resp)


def iter_blob_chunks(data, chunk_size=WRITE_BUFFER_SIZE):
    """Generator for chunks of Blob data.


    The first chunk always bear the Blob oid and the full Blob size.
    The next ones don't have them (i.e., they have the gRPC default values)
    """
    # better than reimplementing our own cursor system
    reader = io.BytesIO(data)

    is_first = True
    while True:
        chunk = reader.read(chunk_size)
        # send first response even if empty
        if is_first or chunk:
            yield chunk, is_first
        else:
            # the break statement would be equivalent, but would appears
            # uncovered. If someone has the courage, it could be a
            # bug to report in coverage 5.5
            return
        is_first = False



# Copyright 2021 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
"""Handling of GitLab refs.

This module is primarily geared towards "special" refs, such as
`refs/merge_requests/1/head`, but should really adopt all general
utilities about refs, i.e, anything about a full ref path, such as
`refs/some/thing` and not yet refined to a given type, such as branches
or tags.
"""
from fnmatch import fnmatch

from heptapod.gitlab.branch import gitlab_branch_ref
from heptapod.gitlab.tag import gitlab_tag_ref
from hgext3rd.heptapod.branch import (
    gitlab_branches,
    get_default_gitlab_branch,
)
from hgext3rd.heptapod.tag import gitlab_tags
from hgext3rd.heptapod import ensure_gitlab_special_refs
from hgext3rd.heptapod import ensure_gitlab_keep_arounds
from hgext3rd.heptapod.special_ref import (
    GITLAB_TYPED_REFS_MISSING,
    parse_special_ref,
    special_refs,
)
from hgext3rd.heptapod.keep_around import (
    iter_keep_arounds,
    KEEP_AROUND_REF_PREFIX,
    KEEP_AROUND_REF_PREFIX_LEN,
)


def gitlab_special_ref_target(repo, ref_path):
    """Return the changeset for a special ref.

    :returns: a :class:`changectx` instance, for the unfiltered version of
       ``repo``. The changeset itself may be obsolete.
    """
    name = parse_special_ref(ref_path)
    if name is None:
        return None

    all_special_refs = special_refs(repo)
    if all_special_refs is GITLAB_TYPED_REFS_MISSING:
        # transitional while we still have an inner Git repo
        # would still be the best we can do near the end of HGitaly2 milestone
        all_special_refs = ensure_special_refs(repo)

    sha = all_special_refs.get(name)
    if sha is None:
        return None

    # TODO catch RepoLookupError (case of inconsistency)
    return repo.unfiltered()[sha]


def path_match(path_segments, pattern_segments):
    # special case for prefix_matching
    if pattern_segments[-1] == b'':
        # direct equality could have issues comparing lists to tuples
        return all(
            path_segment == pat_segment
            for path_segment, pat_segment in zip(path_segments,
                                                 pattern_segments[:-1]))

    return (len(pattern_segments) == len(path_segments)
            and all(
                fnmatch(path_segment, pat_segment)
                for path_segment, pat_segment in zip(path_segments,
                                                     pattern_segments)
                ))


def path_match_any(path, patterns):
    """Tell if path matches at least one of the given glob-like patterns.

    The matching is done as in `git-for-each-ref`, that is either as in
    `fnmatch(3)` (C stdlib function, matches segment by segment of the path)
    or as a direct prefix if pattern ends
    with a slash.

    Because Python's :func:`fnmatch` does not match per slash-separated
    segment, but on the whole string, considering the slash to be an ordinary
    character, we implement the segmentation, and hence need patterns to
    come pre-split at slashes.

    :param patterns: iterable of patterns, each pattern being an iterable
      of segment level patterns, e.g., ``[b'refs', b'heads', b'branch', b'*']``
    """
    path_segments = path.split(b'/')
    return any(path_match(path_segments, pat) for pat in patterns)


def iter_gitlab_special_refs_as_refs(repo, deref=True, patterns=None):
    """Iterate on special refs, yielding full ref paths and target hashes.

    :param patterns: as in :func:`path_match_any` (presplit)
    """
    all_special_refs = special_refs(repo)
    if all_special_refs is GITLAB_TYPED_REFS_MISSING:
        # transitional while we still have an inner Git repo
        # would still be the best we can do near the end of HGitaly2 milestone
        all_special_refs = ensure_special_refs(repo)

    for sref, sha in all_special_refs.items():
        ref_path = b'refs/' + sref
        if patterns and not path_match_any(ref_path, patterns):
            continue
        yield ref_path, (repo[sha] if deref else sha.decode('ascii'))


def ensure_special_refs(repo):
    return ensure_gitlab_special_refs(repo.ui, repo)


def has_keep_around(repo, sha):
    """Tell if there is a keep around for the given changeset hash.

    :param bytes sha: the changeset hash.
    """
    for ka in iter_keep_arounds(repo):
        if ka is GITLAB_TYPED_REFS_MISSING:
            ensure_gitlab_keep_arounds(repo.ui, repo)
            return has_keep_around(repo, sha)
        if ka == sha:
            return True
    return False


def keep_around_ref_path(sha):
    # TODO should move to py-heptapod
    return KEEP_AROUND_REF_PREFIX + sha


def parse_keep_around_ref_path(ref):
    if not ref.startswith(KEEP_AROUND_REF_PREFIX):
        return None
    return ref[KEEP_AROUND_REF_PREFIX_LEN:]


def iter_keep_arounds_as_refs(repo, deref=True, patterns=None):
    for sha in iter_keep_arounds(repo):
        if sha is GITLAB_TYPED_REFS_MISSING:
            ensure_gitlab_keep_arounds(repo.ui, repo)
            yield from iter_keep_arounds_as_refs(repo, deref=deref)
            return
        ref_path = keep_around_ref_path(sha)
        if patterns and not path_match_any(ref_path, patterns):
            continue

        yield (ref_path, (repo[sha] if deref else sha.decode('ascii')))


def sort_key_ref_changeset_date_asc(ref_chgset):
    """Sorting key function meant for ref items.

    :param ref_chgset: a pair made of ref name and changeset (changectx)
    """
    return ref_chgset[1].date()[0]


def sort_key_ref_changeset_date_desc(ref_chgset):
    """Sorting key function meant for ref items.

    :param ref_chgset: a pair made of ref name and changeset (changectx)
    """
    return -ref_chgset[1].date()[0]


def reverse_refs(repo):
    reverse_refs = getattr(repo, '_hgitaly_reverse_refs', None)
    if reverse_refs is not None:
        return reverse_refs

    reverse_refs = {}
    tags = gitlab_tags(repo)
    if tags is not GITLAB_TYPED_REFS_MISSING:
        for tag_name, target in tags.items():
            reverse_refs.setdefault(target, []).append(
                gitlab_tag_ref(tag_name))
    special = special_refs(repo)
    branches = gitlab_branches(repo)
    if branches is not GITLAB_TYPED_REFS_MISSING:
        for branch_name, target in branches.items():
            reverse_refs.setdefault(target, []).append(
                gitlab_branch_ref(branch_name))
    if special is not GITLAB_TYPED_REFS_MISSING:
        for ref_name, target in special.items():
            reverse_refs.setdefault(target, []).append(b'refs/' + ref_name)
    default_branch = get_default_gitlab_branch(repo)
    if (
            default_branch is not None
            and branches is not GITLAB_TYPED_REFS_MISSING
    ):
        target = branches.get(default_branch)
        if target is not None:
            reverse_refs.setdefault(target, []).append(b'HEAD')

    repo._hgitaly_reverse_refs = reverse_refs
    return reverse_refs



# Copyright 2020 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
"""Tests with Gitaly are the tests where we directly compare the Hgitaly
response with Gitaly response. Here, the main motive is to make sure that
Hgitaly responses are what Gitlab expects.

In these tests we use Heptapod's GitLab mirror to write the Git repo.

Lots of duplication from py-heptapod tests and setup reuse from
other HGitaly tests feels weird because asymetrical, but it makes the
point.

We don't need to pre-create the Git repo, because GitLab mirror will
do it automatically.
"""
import os
from pathlib import Path

import hgitaly
import logging
import subprocess

logger = logging.getLogger(__name__)


def gitaly_home():  # pragma no cover
    """Find where Gitaly sits on the file system


    :returns: absolute :class:`Path` instance for the Gitaly source
       checkout, and the relative path to the executable

    The Gitaly sources must have been built.
    This first version assumes we are in HDK context or that
    the ``GITALY_INSTALL_DIR`` environment variable is set (typically in CI).
    """
    from_env = os.environ.get("GITALY_INSTALL_DIR")
    if from_env is not None:
        bin_dir = install_dir = Path(from_env).resolve()
    else:
        hgitaly_clone = Path(hgitaly.__file__).resolve().parent.parent
        hdk = hgitaly_clone.parent
        install_dir = hdk / 'gitaly'
        bin_dir = install_dir / '_build/bin'

    if bin_dir.is_dir():
        gitaly = bin_dir / 'gitaly'
        if gitaly.is_file():
            # startup costs about 0.15s. That's a lot, but it will
            # make sure our skip is correct and it is useful to log the
            # version number
            version = subprocess.check_output((gitaly, '-version')).decode()
            assert 'Gitaly' in version
            logger.info("Found %s at %s. Will lauch comparison tests "
                        "of HGitaly with Gitaly", version, gitaly)
            return install_dir, bin_dir

    return None, None


GITALY_INSTALL_DIR, GITALY_BIN_DIR = gitaly_home()


def gitaly_not_installed():
    return GITALY_INSTALL_DIR is None


def skip_comparison_tests():
    return (gitaly_not_installed()
            or os.environ.get('SKIP_GITALY_COMPARISON_TESTS'))



# Copyright 2021 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
from contextlib import contextmanager
from tempfile import NamedTemporaryFile


@contextmanager
def client_command(priv_key, known_hosts):
    """Context manager providing a suitable ssh client command.

    The command references temporary files that get cleaned up
    automatically when exiting the context.

    :param str priv_key: non-encrypted private key file content.
    """
    with (
            NamedTemporaryFile(mode='w', delete=False) as keyf,
            NamedTemporaryFile(mode='w', delete=False) as known_hostsf,
    ):
        keyf.write(priv_key)
        keyf.close()
        known_hostsf.write(known_hosts)
        known_hostsf.close()

        yield ['ssh',
               '-o', 'IdentitiesOnly=yes',
               '-o', 'UserKnownHostsFile=%s' % known_hostsf.name,
               # do not prompt user if server key is unknown
               '-o', 'StrictHostKeyChecking=yes',
               '-i', keyf.name
               ]


@contextmanager
def client_command_str(*args, **kwargs):
    with client_command(*args, **kwargs) as cmd:
        yield ' '.join(cmd)



# coding: utf-8
# Copyright 2021 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
import logging
import os

from grpc import StatusCode

from mercurial import (
    exchange,
    pycompat,
)

from ..branch import iter_gitlab_branches_matching
from ..peer import (
    FileURLOutsidePath,
    InvalidURLScheme,
    PeerInitException,
    URLParseError,
    hg_remote_peer,
)
from ..repository import (
    config_inherits,
    heptapod_config,
    heptapod_local_config,
    set_config_inheritance,
    set_managed_config,
)
from ..errors import (
    not_implemented,
)
from ..stub.mercurial_repository_pb2 import (
    ConfigItemType,
    GetConfigItemRequest,
    GetConfigItemResponse,
    GetManagedConfigRequest,
    GetManagedConfigResponse,
    PushRequest,
    PushResponse,
    SetManagedConfigRequest,
    SetManagedConfigResponse,
)
from ..stub.mercurial_repository_pb2_grpc import (
    MercurialRepositoryServiceServicer,
)
from ..servicer import HGitalyServicer

logger = logging.getLogger(__name__)


class MercurialRepositoryServicer(MercurialRepositoryServiceServicer,
                                  HGitalyServicer):
    """MercurialRepositoryService implementation.

    The ordering of methods in this source file is the same as in the proto
    file.
    """
    def GetConfigItem(self,
                      request: GetConfigItemRequest,
                      context) -> GetConfigItemResponse:
        repo = self.load_repo(request.repository, context)
        section = pycompat.sysbytes(request.section)
        name = pycompat.sysbytes(request.name)

        if request.as_type == ConfigItemType.BOOL:
            # TODO error treatment if value is not boolean
            return GetConfigItemResponse(
                as_bool=repo.ui.configbool(section, name))

        not_implemented(context, issue=60)  # pragma no cover

    def GetManagedConfig(self,
                         request: GetManagedConfigRequest,
                         context) -> GetManagedConfigResponse:
        repo = self.load_repo(request.repository, context)
        if request.local:
            heptapod_section = heptapod_local_config(repo)
        else:
            heptapod_section = heptapod_config(repo)
        return GetManagedConfigResponse(inherit=config_inherits(repo),
                                        heptapod=heptapod_section)

    def SetManagedConfig(self,
                         request: SetManagedConfigRequest,
                         context) -> SetManagedConfigResponse:
        repo = self.load_repo(request.repository, context)
        set_managed_config(repo,
                           heptapod=request.heptapod,
                           remove_items=request.remove_items,
                           by_line=request.by_line)
        if request.HasField('inherit'):
            if not request.inherit:
                hgrc_dir = None  # remove
            else:
                project_path = request.repository.gl_project_path
                group_path = project_path.rsplit('/', 1)[0]
                hgrc_dir = os.path.relpath(
                    group_path,
                    request.repository.relative_path + '/.hg'
                )
            set_config_inheritance(repo, hgrc_dir, request.by_line)

        return SetManagedConfigResponse()

    def Push(self, request: PushRequest, context) -> PushResponse:
        repo = self.load_repo(request.repository, context)
        repo.ui.setconfig(b'hooks', b'pretxnclose.heptapod_sync', b'')
        repo.ui.setconfig(b'experimental', b'auto-publish', b'abort')

        remote_url = request.remote_peer.url

        # storage name has already been validated by the repository resolution
        storage_path = os.fsdecode(
            self.storages.get(request.repository.storage_name))

        try:
            with hg_remote_peer(repo, request.remote_peer,
                                storage_path=storage_path) as remote_peer:
                include_drafts = request.include_drafts
                only_revs = None
                only_branches_matching = request.only_gitlab_branches_matching
                if only_branches_matching:
                    only_heads = [
                        ctx.hex()
                        for _name, ctx in iter_gitlab_branches_matching(
                            repo, only_branches_matching)
                    ]
                    if include_drafts:
                        only_revs = repo.revs('%ls', only_heads)
                    else:
                        # Note that this is public ancestors of the given
                        # heads,  not the same as ancestors of public heads.
                        # This form is preferred because it doesn't block the
                        # push of the protected branch if its head happens
                        # to be a draft.
                        only_revs = repo.revs('public() and ::(%ls)',
                                              only_heads)
                elif not include_drafts:
                    only_revs = repo.revs('public()')

                if only_revs is not None and not only_revs:
                    return PushResponse(new_changesets=False)

                push_kwargs = dict(newbranch=True)
                if only_revs is not None:
                    push_kwargs['revs'] = [repo[rev].node()
                                           for rev in only_revs]

                try:
                    pushop = exchange.push(repo=repo, remote=remote_peer,
                                           **push_kwargs)
                except Exception as exc:
                    context.abort(StatusCode.INTERNAL,
                                  "Error pushing to %r: %s" % (remote_url,
                                                               exc))

                new_changesets = (pushop.cgresult != 0
                                  and pushop.cgresult is not None)
                return PushResponse(new_changesets=new_changesets)
        except URLParseError as exc:
            context.abort(StatusCode.INVALID_ARGUMENT,
                          "Invalid URL %r for push: %s " % exc.args)
        except InvalidURLScheme as exc:
            context.abort(StatusCode.INVALID_ARGUMENT,
                          "Invalid scheme %r for push URL %r. "
                          "Please use one of %r" % (exc.args[0],
                                                    remote_url, exc.args[1]))
        except FileURLOutsidePath as exc:
            context.abort(StatusCode.INVALID_ARGUMENT,
                          "file URL %r not under storage %r "
                          "root directory" % exc.args)
        except PeerInitException as exc:
            context.abort(StatusCode.INVALID_ARGUMENT,
                          "URL %r could not be used as "
                          "a peer for push: %s" % (remote_url, exc))
        except Exception as exc:
            code = context.code()
            if code is not None and code != StatusCode.OK:
                # already treated: context.abort() raises Exception() (sic)
                raise

            context.abort(StatusCode.INTERNAL,
                          "Unexpected error, not in the actual "
                          "push to %r: %s" % (remote_url, exc))



# Copyright 2021 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later

import pytest
from heptapod.testhelpers import (
    LocalRepoWrapper,
)
from heptapod.testhelpers.gitlab import GitLabMirrorFixture

from hgext3rd.heptapod.branch import read_gitlab_typed_refs
from hgext3rd.heptapod.special_ref import (
    write_gitlab_special_ref,
)
from hgext3rd.heptapod.keep_around import (
    create_keep_around,
)
from .common import (
    MINIMAL_HG_CONFIG,
)
from ..gitlab_ref import (
    gitlab_special_ref_target,
    iter_gitlab_special_refs_as_refs,
    iter_keep_arounds_as_refs,
    keep_around_ref_path,
    parse_keep_around_ref_path,
)


@pytest.fixture
def repo_wrapper(tmpdir):
    wrapper = LocalRepoWrapper.init(tmpdir, config=MINIMAL_HG_CONFIG)
    yield wrapper.repo, wrapper


@pytest.fixture
def mirror(tmpdir, monkeypatch):
    config = MINIMAL_HG_CONFIG.copy()
    config['extensions']['heptapod'] = ''

    with GitLabMirrorFixture.init(tmpdir, monkeypatch,
                                  common_repo_name='repo',
                                  hg_config=config) as mirror:
        mirror.activate_mirror()
        yield mirror


def test_gitlab_special_ref_target(repo_wrapper):
    repo, wrapper = repo_wrapper

    ref_name = b'merge-requests/1/head'
    ref_path = b'refs/merge-requests/1/head'

    # empty repo, the file doesn't even exist
    assert gitlab_special_ref_target(repo, ref_path) is None

    base = wrapper.commit_file('foo')
    write_gitlab_special_ref(repo, ref_name, base)

    assert gitlab_special_ref_target(repo, ref_path) == base

    # making target obsolete doesn't hide it to the special refs subsystem
    successor = wrapper.amend_file('foo')
    assert gitlab_special_ref_target(repo, ref_path) == base

    # updates are applied immediately (cache is updated)
    write_gitlab_special_ref(repo, ref_name, successor.hex())
    assert gitlab_special_ref_target(repo, ref_path) == successor

    # unknown, not special, alien and completely bogus cases
    assert gitlab_special_ref_target(repo, b'refs/pipelines/123') is None
    assert gitlab_special_ref_target(repo, b'refs/heads/branch/main') is None
    assert gitlab_special_ref_target(repo, b'refs/pull/123/head') is None
    assert gitlab_special_ref_target(repo, b'bogus') is None


def test_gitlab_special_ref_target_ensure(mirror):
    wrapper = mirror.hg_repo_wrapper
    git_repo = mirror.git_repo

    ctx = wrapper.commit_file('foo')
    git_sha = git_repo.branches()[b'branch/default']['sha']

    ref_path = b'refs/environments/654'
    git_repo.write_ref(ref_path, git_sha)

    assert gitlab_special_ref_target(wrapper.repo, ref_path) == ctx


def test_write_special_ref(repo_wrapper):
    repo, wrapper = repo_wrapper

    ref_name = b'pipelines/123'
    ref_path = b'refs/pipelines/123'

    base = wrapper.commit_file('foo')
    write_gitlab_special_ref(repo, ref_name, base)

    # direct read without cache
    assert read_gitlab_typed_refs(repo, 'special-refs') == {
        ref_name: base.hex()}

    # cache got updated (actually, created) anyway
    assert gitlab_special_ref_target(repo, ref_path) == base

    # passing a hex sha (bytes) also works and cache is updated
    ctx1 = wrapper.commit_file('foo')
    write_gitlab_special_ref(repo, ref_name, ctx1.hex())
    assert read_gitlab_typed_refs(repo, 'special-refs') == {
        ref_name: ctx1.hex()}
    assert gitlab_special_ref_target(repo, ref_path) == ctx1


def test_iterate_special_ref(repo_wrapper):
    repo, wrapper = repo_wrapper

    ref_name = b'pipelines/123'
    ref_path = b'refs/pipelines/123'

    base = wrapper.commit_file('foo')
    base_hex = base.hex().decode()

    assert list(iter_gitlab_special_refs_as_refs(repo)) == []
    write_gitlab_special_ref(repo, ref_name, base)

    assert list(iter_gitlab_special_refs_as_refs(repo)) == [
        (ref_path, base)]
    assert list(iter_gitlab_special_refs_as_refs(repo, deref=False)) == [
        (ref_path, base_hex)]


def test_iterate_keep_arounds(repo_wrapper):
    repo, wrapper = repo_wrapper
    ctx = wrapper.commit_file('foo')
    sha_bytes = ctx.hex()
    sha_str = sha_bytes.decode()

    assert list(iter_keep_arounds_as_refs(repo)) == []
    create_keep_around(repo, sha_bytes)

    assert list(iter_keep_arounds_as_refs(repo)) == [
        (b'refs/keep-around/' + sha_bytes, ctx)]
    assert list(iter_keep_arounds_as_refs(repo, deref=False)) == [
        (b'refs/keep-around/' + sha_bytes, sha_str)]


def test_parse_keep_around_ref_path():
    sha = b'12abdc43' * 5
    ka_ref = b'refs/keep-around/' + sha
    assert parse_keep_around_ref_path(ka_ref) == sha
    assert keep_around_ref_path(sha) == ka_ref

    assert parse_keep_around_ref_path(b'refs/pipeline/17') is None



# Copyright 2023 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
from google.protobuf.timestamp_pb2 import Timestamp
from grpc import (
    RpcError,
    StatusCode,
)
import pytest

from mercurial import (
    phases,
)

from heptapod.testhelpers.git import GitRepo
from hgext3rd.heptapod.branch import (
    gitlab_branches,
)

from hgitaly.stub.operations_pb2 import (
    OperationBranchUpdate,
    UserFFBranchRequest,
    UserFFBranchResponse,
    UserSquashRequest,
)
from hgitaly.stub.operations_pb2_grpc import (
    OperationServiceStub,
)
from hgitaly.stub.shared_pb2 import (
    User,
)
from hgitaly.tests.common import make_empty_repo_with_gitlab_state_maintainer

from .fixture import MutationServiceFixture

parametrize = pytest.mark.parametrize


class OperationsFixture(MutationServiceFixture):

    stub_cls = OperationServiceStub

    def user_squash(self, **kw):
        kw.setdefault('repository', self.grpc_repo)
        kw.setdefault('user', self.user)
        return self.stub.UserSquash(UserSquashRequest(**kw),
                                    metadata=self.grpc_metadata())

    def user_ff_branch(self, **kw):
        kw.setdefault('repository', self.grpc_repo)
        kw.setdefault('user', self.user)
        return self.stub.UserFFBranch(UserFFBranchRequest(**kw),
                                      metadata=self.grpc_metadata())


@pytest.fixture
def operations_fixture(grpc_channel, server_repos_root):
    with OperationsFixture(
            grpc_channel, server_repos_root,
            repo_factory=make_empty_repo_with_gitlab_state_maintainer
    ) as fixture:
        yield fixture


@parametrize('timestamp', ('timestamp', 'now'))
@parametrize('project_mode', ('hg-git-project',
                              'native-project-without-git',
                              'native-project-with-git',
                              ))
def test_user_squash(operations_fixture, project_mode, timestamp):
    fixture = operations_fixture
    wrapper = fixture.repo_wrapper
    hg_repo = wrapper.repo

    # because of the config set by fixture, this leads in all cases to
    # creation of a Git repo and its `branch/default` Git branch
    sha1 = wrapper.commit_file('foo').hex().decode('ascii')
    sha2 = wrapper.commit_file('foo', message='foo2').hex().decode('ascii')
    sha3 = wrapper.commit_file('foo', message='foo3').hex().decode('ascii')
    wrapper.update(sha1)  # avoid keeping changeset 3 visible
    # let's confirm it
    git_repo = GitRepo(str(wrapper.path) + '.git')
    before_squash_git_branches = git_repo.branches()
    assert before_squash_git_branches[b'branch/default']['title'] == b'foo3'

    squash = fixture.user_squash

    fixture.hg_native = project_mode != 'hg-git-project'
    if fixture.hg_native:
        fixture.with_hg_git_mirroring = (
            project_mode == 'native-project-with-git'
        )

    if timestamp == 'now':
        ts = None
    else:
        ts = Timestamp()
        ts.FromSeconds(1702472217)
    author = User(name=b"John Doe",
                  gl_id='user-987',
                  email=b"jd@heptapod.test",
                  )
    resp = squash(start_sha=sha1,
                  end_sha=sha3,
                  author=author,
                  timestamp=ts,
                  commit_message=b'squashed!')

    # we will need more that list_refs, namely that the state file does exist,
    # without any fallback.
    # TODO check obslog on the result
    wrapper.reload()
    hg_repo = wrapper.repo
    gl_branches = gitlab_branches(hg_repo)
    folded_sha = gl_branches[b'branch/default']
    assert folded_sha == resp.squash_sha.encode('ascii')
    assert folded_sha != sha2.encode('ascii')
    folded_ctx = hg_repo[folded_sha]
    assert folded_ctx.description() == b'squashed!'
    unfi = hg_repo.unfiltered()
    for sha in (sha2, sha3):
        assert unfi[sha].obsolete()
    assert fixture.list_refs() == {
        b'refs/heads/branch/default': folded_sha.decode('ascii')
    }

    if fixture.hg_native and not fixture.with_hg_git_mirroring:
        # expect Git branches not to have moved
        assert git_repo.branches() == before_squash_git_branches
    else:
        # expect Git branch to point on new commit
        assert git_repo.branches()[b'branch/default']['title'] == b'squashed!'

    for kw in (
            dict(user=None, start_sha=sha1, end_sha=sha2),
            dict(commit_message=None, start_sha=sha1, end_sha=sha2),
            # missing author
            dict(commit_message=b'squashed', start_sha=sha1, end_sha=sha2),
            # missing start_sha or end_sha
            dict(commit_message=b'squashed', author=author, end_sha=sha2),
            dict(commit_message=b'squashed', author=author, start_sha=sha1),
            dict(commit_message=b'squashed', author=author,
                 start_sha='unknown', end_sha=sha2),
            dict(commit_message=b'squashed', author=author, end_sha=sha2),
            dict(commit_message=b'squashed', author=author,
                 start_sha=sha1, end_sha='unknown'),
    ):
        with pytest.raises(RpcError) as exc_info:
            squash(**kw)
        assert exc_info.value.code() == StatusCode.INVALID_ARGUMENT


@parametrize('project_mode', ('hg-git-project',
                              'native-project-without-git',
                              'native-project-with-git',
                              ))
def test_user_ff_branch(operations_fixture, project_mode):
    fixture = operations_fixture
    ff_branch = fixture.user_ff_branch

    fixture.hg_native = project_mode != 'hg-git-project'
    if fixture.hg_native:
        fixture.with_hg_git_mirroring = (
            project_mode == 'native-project-with-git'
        )
    wrapper = fixture.repo_wrapper

    gl_topic = b'topic/default/zetop'
    gl_branch = b'branch/default'

    # because of the config set by fixture, this leads in all cases to
    # creation of a Git repo and its `branch/default` Git branch
    ctx0 = wrapper.commit_file('foo')
    sha0 = ctx0.hex().decode()
    default_head = wrapper.commit_file('foo')
    ctx2 = wrapper.commit_file('foo', topic='zetop', message='foo2')
    sha2 = ctx2.hex().decode('ascii')
    wrapper.commit_file('bar', branch='other', parent=ctx0)
    needs_rebase = wrapper.commit_file('old', parent=ctx0,
                                       topic='needs-rebase'
                                       ).hex().decode('ascii')
    bogus1 = wrapper.commit_file('bogus', topic='bogus',
                                 parent=default_head)
    bogus1_sha = bogus1.hex().decode()
    # changing topic so that amending bogus1 is not a multiple heads condition
    bogus2 = wrapper.commit_file('bogus', topic='bogus2')
    bogus2_sha = bogus2.hex().decode()

    # let's confirm the mirroring to Git
    git_repo = fixture.git_repo()
    before_ff_git_branches = git_repo.branches()
    assert before_ff_git_branches[gl_topic]['title'] == b'foo2'

    # making bogus_top1 obsolete
    wrapper.update_bin(bogus1.node())
    wrapper.amend_file('foo')

    before_refs = fixture.list_refs()

    with pytest.raises(RpcError) as exc_info:
        ff_branch(branch=b'branch/other', commit_id=sha2)
    assert exc_info.value.code() == StatusCode.FAILED_PRECONDITION
    assert 'branch differ' in exc_info.value.details()

    with pytest.raises(RpcError) as exc_info:
        ff_branch(branch=gl_branch, commit_id=bogus2_sha)
    assert exc_info.value.code() == StatusCode.FAILED_PRECONDITION
    assert 'unstable' in exc_info.value.details()

    with pytest.raises(RpcError) as exc_info:
        ff_branch(branch=gl_branch, commit_id=bogus1_sha)
    assert exc_info.value.code() == StatusCode.FAILED_PRECONDITION
    assert 'obsolete' in exc_info.value.details()

    with pytest.raises(RpcError) as exc_info:
        ff_branch(branch=gl_topic, commit_id=sha2)
    assert exc_info.value.code() == StatusCode.FAILED_PRECONDITION
    assert 'named branches only' in exc_info.value.details()

    # case where nothing is pathological, but is not a fast-forward
    with pytest.raises(RpcError) as exc_info:
        ff_branch(branch=gl_branch, commit_id=needs_rebase)
    assert exc_info.value.code() == StatusCode.FAILED_PRECONDITION
    assert 'not fast forward' in exc_info.value.details()

    # basic errors, missing and unresolvable arguments
    with pytest.raises(RpcError) as exc_info:
        ff_branch(commit_id=sha2)
    assert exc_info.value.code() == StatusCode.INVALID_ARGUMENT
    assert 'empty branch' in exc_info.value.details()

    with pytest.raises(RpcError) as exc_info:
        ff_branch(branch=gl_branch, commit_id='not-a-hash')
    assert exc_info.value.code() == StatusCode.INVALID_ARGUMENT
    assert 'parse commit ID' in exc_info.value.details()

    unknown_oid = '01beef23' * 5
    with pytest.raises(RpcError) as exc_info:
        ff_branch(branch=gl_branch, commit_id=unknown_oid)
    assert exc_info.value.code() == StatusCode.INTERNAL
    assert 'invalid commit' in exc_info.value.details()

    with pytest.raises(RpcError) as exc_info:
        ff_branch(branch=gl_branch, commit_id=sha2,
                  expected_old_oid=unknown_oid)
    assert exc_info.value.code() == StatusCode.INVALID_ARGUMENT
    assert 'old object' in exc_info.value.details()

    with pytest.raises(RpcError) as exc_info:
        ff_branch(branch=gl_branch, commit_id=sha2,
                  expected_old_oid='12deadbeef')  # short hash
    assert exc_info.value.code() == StatusCode.INVALID_ARGUMENT
    assert 'parse commit ID' in exc_info.value.details()

    # old oid mismatch
    assert ff_branch(branch=gl_branch, commit_id=sha2,
                     expected_old_oid=sha0) == UserFFBranchResponse()
    assert fixture.list_refs() == before_refs

    # Actual call expected to succeed
    resp = ff_branch(branch=gl_branch, commit_id=sha2)

    # checking change on default branch ref, and only this ref.
    expected_refs = before_refs.copy()
    expected_refs[b'refs/heads/' + gl_branch] = sha2
    del expected_refs[b'refs/heads/' + gl_topic]

    assert fixture.list_refs() == expected_refs
    assert resp.branch_update == OperationBranchUpdate(commit_id=sha2,
                                                       repo_created=False,
                                                       branch_created=False)
    wrapper.reload()
    assert wrapper.repo[ctx2.rev()].phase() == phases.public



# Copyright 2023 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
import attr
import contextlib
import logging
import random
import pytest
import shutil

import grpc

from heptapod.testhelpers.hg import RepoWrapper
from heptapod.testhelpers.gitlab import patch_gitlab_hooks

from hgitaly.stub.shared_pb2 import Repository
from .comparison import BaseRpcHelper


logger = logging.getLogger(__name__)


@attr.s
class HGitalyRHGitalyComparison():
    """This fixture is for comparison between HGitaly and RHGitaly.

    As an edge case, it can be used to test RHGitaly alone, hence the
    `hgitaly_channel` is optional.

    Gitaly itself is not involved at all, so this is for HGitaly-specific
    calls, often but not necessarily involving the Mercurial services.
    """
    rhgitaly_channel = attr.ib()

    hgitaly_channel = attr.ib(default=None)
    hgitaly_repo = attr.ib(default=None)
    hg_repo_wrapper = attr.ib(default=None)
    gitlab_sent_hooks = attr.ib(default=attr.Factory(list))

    def activate_gitlab_state_maintainer(self, monkeypatch):
        # TODO expose that from py-heptapod
        self.hg_repo_wrapper.repo.ui.setconfig(
            b'hooks', b'pretxnclose.testcase',
            b'python:heptapod.hooks.gitlab_mirror.mirror')
        patch_gitlab_hooks(monkeypatch, self.gitlab_sent_hooks)

    def rpc_helper(self, **kw):
        return RpcHelper(self, **kw)

    def invalidate(self):  # pragma no cover
        """Invalidate all caches.

        In particular, reload the Mercurial repo
        """
        wrapper = self.hg_repo_wrapper
        if wrapper is not None:
            wrapper.reload()

    @property
    def gitaly_repo(self):
        """Compatibility wrapper

        RPC helpers and the like usually expect a `gitaly_repo` attribute.
        """
        return self.hgitaly_repo


class RpcHelper(BaseRpcHelper):
    """Encapsulates a comparison fixture with call and compare helpers.

    This is derived from :class:`hgitaly.comparison.RpcHelper`, but it is
    very much simpler.
    """
    error_details_normalizer = None

    def init_stubs(self):
        comparison, stub_cls = self.comparison, self.stub_cls
        self.stubs = dict(rhgitaly=stub_cls(comparison.rhgitaly_channel))
        hgitaly_channel = comparison.hgitaly_channel
        if hgitaly_channel is not None:
            self.stubs['hgitaly'] = stub_cls(comparison.hgitaly_channel)

    def assert_compare(self, **kwargs):
        self.apply_request_defaults(kwargs)
        assert self.rpc('hgitaly', **kwargs) == self.rpc('rhgitaly', **kwargs)

    def assert_compare_errors(self, same_details=True,
                              skip_structured_errors_issue=None,
                              structured_errors_handler=None,
                              **hg_kwargs):
        """Compare errors returned by (R)HGitaly and Gitaly.

        :param:`structured_errors_handler`: if supplied, it is expected to be
          a :class:`dict` with the following keys:

          - `git_cls`: expected error class (gRPC message) from Gitaly
          - `hg_cls`: expected error class (gRPC message) from HGitaly
                      defaults to the value of `git_cls`
          - `to_git`: conversion callable from HGitaly's gRPC error message
            to Gitaly's
          - `git_normalizer`: additional normalizer to apply to Git error
            (some of them involve objects that cannot exist on the Mercurial
             side)
        """
        self.apply_request_defaults(hg_kwargs)

        with pytest.raises(grpc.RpcError) as exc_info_hg:
            self.rpc('hgitaly', **hg_kwargs)
        with pytest.raises(grpc.RpcError) as exc_info_git:
            self.rpc('rhgitaly', **hg_kwargs)
        self.assert_compare_grpc_exceptions(
            exc_info_hg.value,
            exc_info_git.value,
            same_details=same_details,
            vcses=('hg', 'hg'),
            skip_structured_errors_issue=skip_structured_errors_issue,
            structured_errors_handler=structured_errors_handler,
        )


@contextlib.contextmanager
def hgitaly_rhgitaly_comparison_fixture(server_repos_root,
                                        rhgitaly_channel,
                                        monkeypatch=None,
                                        hgitaly_channel=None,
                                        skip_repo=False,
                                        ):
    if skip_repo:
        hgitaly_repo = hg_repo_wrapper = None
    else:
        relative_path = 'repo-' + hex(random.getrandbits(64))[2:] + '.hg'
        storage = 'default'

        hgitaly_repo = Repository(relative_path=relative_path,
                                  storage_name=storage)

        hg_config = dict(
            phases=dict(publish=False),
            ui=dict(username='Hgitaly Tests <hgitaly@heptapod.test>'),
            extensions={name: '' for name in ('evolve',
                                              'topic',
                                              'hgitaly',
                                              'heptapod')})
        hg_repo_wrapper = RepoWrapper.init(
            server_repos_root / storage / relative_path,
            config=hg_config)

    try:
        hrc = HGitalyRHGitalyComparison(rhgitaly_channel=rhgitaly_channel,
                                        hgitaly_channel=hgitaly_channel,
                                        hgitaly_repo=hgitaly_repo,
                                        hg_repo_wrapper=hg_repo_wrapper,
                                        )
        if not skip_repo:
            hrc.activate_gitlab_state_maintainer(monkeypatch)
        yield hrc

    finally:
        if hg_repo_wrapper is not None:
            hg_path = hg_repo_wrapper.path
            try:
                shutil.rmtree(hg_path)
            except Exception:  # pragma no cover
                logger.exception("Error removing the Mercurial repo at %r",
                                 hg_path)
